@@ -16,165 +16,158 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#ifdef JAVA
-#include "javadefines.h"
-#define MdbBackendType_STRUCT_ELEMENT(a,b,c,d) new MdbBackendType(a,b,c,d)
-#else
-#define MdbBackendType_STRUCT_ELEMENT(a,b,c,d) {a,b,c,d}
 /*
 ** functions to deal with different backend database engines
 */
-#
+
 #include "mdbtools.h"
 
-#ifdef DMALLOC
-#include "dmalloc.h"
-#endif
-
-#endif   /* JAVA */
-
-static int is_init;
-GHashTable *mdb_backends;
-void _mdb_remove_backends();
-
 /*    Access data types */
-static MdbBackendType mdb_access_types[] = {
-		MdbBackendType_STRUCT_ELEMENT("Unknown 0x00", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Boolean", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Byte", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Integer", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Long Integer", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Currency", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Single", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Double", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("DateTime", 0,0,1),
-		MdbBackendType_STRUCT_ELEMENT("Binary", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Text", 1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("OLE", 1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("Memo/Hyperlink",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("Unknown 0x0d",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Unknown 0x0e",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Replication ID",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Numeric",1,1,0)
+static const MdbBackendType mdb_access_types[] = {
+    { .name = "Unknown 0x00" },
+    { .name = "Boolean" },
+    { .name = "Byte" },
+    { .name = "Integer" },
+    { .name = "Long Integer" },
+    { .name = "Currency" },
+    { .name = "Single" },
+    { .name = "Double" },
+    { .name = "DateTime", .needs_quotes = 1 },
+    { .name = "Binary" },
+    { .name = "Text", .needs_length = 1, .needs_quotes = 1 },
+    { .name = "OLE", .needs_length = 1, .needs_quotes = 1 },
+    { .name = "Memo/Hyperlink", .needs_length = 1, .needs_quotes = 1 },
+    { .name = "Unknown 0x0d" },
+    { .name = "Unknown 0x0e" },
+    { .name = "Replication ID" },
+    { .name = "Numeric", .needs_length = 1, .needs_scale = 1 },
 };
 
 /*    Oracle data types */
-static MdbBackendType mdb_oracle_types[] = {
-		MdbBackendType_STRUCT_ELEMENT("Oracle_Unknown 0x00",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("NUMBER(1)",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("NUMBER(3)",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("NUMBER(5)",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("NUMBER(11)",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("NUMBER(15,2)",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("FLOAT",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("FLOAT",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("TIMESTAMP",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("BINARY",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("VARCHAR2",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("BLOB",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("CLOB",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("Oracle_Unknown 0x0d",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Oracle_Unknown 0x0e",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("NUMBER",1,0,0),
-		MdbBackendType_STRUCT_ELEMENT("NUMBER",1,0,0),
+static const MdbBackendType mdb_oracle_types[] = {
+    { .name = "Oracle_Unknown 0x00" },
+    { .name = "NUMBER(1)" },
+    { .name = "NUMBER(3)" },
+    { .name = "NUMBER(5)" },
+    { .name = "NUMBER(11)" },
+    { .name = "NUMBER(15,2)" },
+    { .name = "FLOAT" },
+    { .name = "FLOAT" },
+    { .name = "TIMESTAMP" },
+    { .name = "BINARY" },
+    { .name = "VARCHAR2", .needs_length = 1, .needs_quotes = 1 },
+    { .name = "BLOB" },
+    { .name = "CLOB" },
+    { .name = "Oracle_Unknown 0x0d" },
+    { .name = "Oracle_Unknown 0x0e" },
+    { .name = "NUMBER", .needs_length = 1 },
+    { .name = "NUMBER", .needs_length = 1 },
 };
-static MdbBackendType mdb_oracle_shortdate_type =
-		MdbBackendType_STRUCT_ELEMENT("DATE",0,0,0);
+static const MdbBackendType mdb_oracle_shortdate_type =
+		    { .name = "DATE" };
 
 /*    Sybase/MSSQL data types */
-static MdbBackendType mdb_sybase_types[] = {
-		MdbBackendType_STRUCT_ELEMENT("Sybase_Unknown 0x00",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("bit",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("char",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("smallint",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("int",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("money",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("real",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("float",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("smalldatetime",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Sybase_Unknown 0x09",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("varchar",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("varbinary",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("text",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("Sybase_Unknown 0x0d",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Sybase_Unknown 0x0e",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Sybase_Replication ID",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("numeric",1,1,0),
+static const MdbBackendType mdb_sybase_types[] = {
+    { .name = "Sybase_Unknown 0x00" },
+    { .name = "bit" },
+    { .name = "char", .needs_length = 1, .needs_quotes = 1 },
+    { .name = "smallint" },
+    { .name = "int" },
+    { .name = "money" },
+    { .name = "real" },
+    { .name = "float" },
+    { .name = "smalldatetime" },
+    { .name = "Sybase_Unknown 0x09" },
+    { .name = "varchar", .needs_length = 1, .needs_quotes = 1 },
+    { .name = "varbinary", .needs_length = 1, .needs_quotes = 1 },
+    { .name = "text", .needs_length = 1, .needs_quotes = 1 },
+    { .name = "Sybase_Unknown 0x0d" },
+    { .name = "Sybase_Unknown 0x0e" },
+    { .name = "Sybase_Replication ID" },
+    { .name = "numeric", .needs_length = 1, .needs_scale = 1 },
 };
-static MdbBackendType mdb_sybase_shortdate_type =
-		MdbBackendType_STRUCT_ELEMENT("DATE",0,0,0);
+static const MdbBackendType mdb_sybase_shortdate_type =
+		    { .name = "DATE" };
 
 /*    Postgres data types */
-static MdbBackendType mdb_postgres_types[] = {
-		MdbBackendType_STRUCT_ELEMENT("Postgres_Unknown 0x00",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("BOOL",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("SMALLINT",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("INTEGER",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("INTEGER",0,0,0), /* bigint */
-		MdbBackendType_STRUCT_ELEMENT("NUMERIC(15,2)",0,0,0), /* money deprecated ? */
-		MdbBackendType_STRUCT_ELEMENT("REAL",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("DOUBLE PRECISION",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("TIMESTAMP WITHOUT TIME ZONE",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("BYTEA",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("VARCHAR",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("BYTEA",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("TEXT",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Postgres_Unknown 0x0d",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Postgres_Unknown 0x0e",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("UUID",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("Postgres_Unknown 0x10",0,0,0),
+static const MdbBackendType mdb_postgres_types[] = {
+    { .name = "Postgres_Unknown 0x00" },
+    { .name = "BOOL" },
+    { .name = "SMALLINT" },
+    { .name = "INTEGER" },
+    { .name = "INTEGER" }, /* bigint */
+    { .name = "NUMERIC(15,2)" }, /* money deprecated ? */
+    { .name = "REAL" },
+    { .name = "DOUBLE PRECISION" },
+    { .name = "TIMESTAMP WITHOUT TIME ZONE" },
+    { .name = "BYTEA" },
+    { .name = "VARCHAR", .needs_length = 1, .needs_quotes = 1 },
+    { .name = "BYTEA" },
+    { .name = "TEXT" },
+    { .name = "Postgres_Unknown 0x0d" },
+    { .name = "Postgres_Unknown 0x0e" },
+    { .name = "UUID" },
+    { .name = "NUMERIC", .needs_length = 1, .needs_scale = 1 },
 };
-static MdbBackendType mdb_postgres_shortdate_type =
-		MdbBackendType_STRUCT_ELEMENT("DATE",0,0,0);
-static MdbBackendType mdb_postgres_serial_type =
-		MdbBackendType_STRUCT_ELEMENT("SERIAL",0,0,0);
+static const MdbBackendType mdb_postgres_shortdate_type =
+		    { .name = "DATE" };
+static const MdbBackendType mdb_postgres_serial_type =
+		    { .name = "SERIAL" };
 
 /*    MySQL data types */
-static MdbBackendType mdb_mysql_types[] = {
-		MdbBackendType_STRUCT_ELEMENT("Text",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("char",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("int",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("int",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("int",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("float",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("float",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("float",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("datetime",0,0,1),
-		MdbBackendType_STRUCT_ELEMENT("varchar",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("varchar",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("varchar",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("text",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("blob",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("text",1,0,1),
-		MdbBackendType_STRUCT_ELEMENT("char(38)",0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("numeric",1,1,0),
+static const MdbBackendType mdb_mysql_types[] = {
+    { .name = "text", .needs_quotes = 1 },
+    { .name = "boolean" },
+    { .name = "tinyint" },
+    { .name = "smallint" },
+    { .name = "int" },
+    { .name = "float" },
+    { .name = "float" },
+    { .name = "float" },
+    { .name = "datetime", .needs_quotes = 1 },
+    { .name = "varchar", .needs_length = 1, .needs_quotes = 1 },
+    { .name = "varchar", .needs_length = 1, .needs_quotes = 1 },
+    { .name = "varchar", .needs_length = 1, .needs_quotes = 1 },
+    { .name = "text", .needs_quotes = 1 },
+    { .name = "blob" },
+    { .name = "text", .needs_quotes = 1 },
+    { .name = "char(38)" },
+    { .name = "numeric", .needs_length = 1, .needs_scale = 1 },
 };
-static MdbBackendType mdb_mysql_shortdate_type =
-		MdbBackendType_STRUCT_ELEMENT("date",0,0,0);
+static const MdbBackendType mdb_mysql_shortdate_type =
+		    { .name = "date" };
 
 /*    sqlite data types */
-static MdbBackendType mdb_sqlite_types[] = {
-		MdbBackendType_STRUCT_ELEMENT("BLOB", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("INTEGER", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("INTEGER", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("INTEGER", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("INTEGER", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("REAL", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("REAL", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("REAL", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("DateTime", 0,0,1),
-		MdbBackendType_STRUCT_ELEMENT("BLOB", 0,0,1),
-		MdbBackendType_STRUCT_ELEMENT("varchar", 0,0,1),
-		MdbBackendType_STRUCT_ELEMENT("BLOB", 0,0,1),
-		MdbBackendType_STRUCT_ELEMENT("TEXT", 0,0,1),
-		MdbBackendType_STRUCT_ELEMENT("BLOB", 0,0,1),
-		MdbBackendType_STRUCT_ELEMENT("BLOB", 0,0,1),
-		MdbBackendType_STRUCT_ELEMENT("INTEGER", 0,0,0),
-		MdbBackendType_STRUCT_ELEMENT("INTEGER", 0,0,0),
+static const MdbBackendType mdb_sqlite_types[] = {
+    { .name = "BLOB" },
+    { .name = "INTEGER" },
+    { .name = "INTEGER" },
+    { .name = "INTEGER" },
+    { .name = "INTEGER" },
+    { .name = "REAL" },
+    { .name = "REAL" },
+    { .name = "REAL" },
+    { .name = "DateTime", .needs_quotes = 1 },
+    { .name = "BLOB", .needs_quotes = 1 },
+    { .name = "varchar", .needs_quotes = 1 },
+    { .name = "BLOB", .needs_quotes = 1 },
+    { .name = "TEXT", .needs_quotes = 1 },
+    { .name = "BLOB", .needs_quotes = 1 },
+    { .name = "BLOB", .needs_quotes = 1 },
+    { .name = "INTEGER" },
+    { .name = "INTEGER" },
 };
 
-#ifndef JAVA
-static gboolean mdb_drop_backend(gpointer key, gpointer value, gpointer data);
+enum {
+	MDB_BACKEND_ACCESS = 1,
+	MDB_BACKEND_ORACLE,
+	MDB_BACKEND_SYBASE,
+	MDB_BACKEND_POSTGRES,
+	MDB_BACKEND_MYSQL,
+	MDB_BACKEND_SQLITE,
+};
+
+static void mdb_drop_backend(gpointer key, gpointer value, gpointer data);
 
 static gchar*
 quote_generic(const gchar *value, gchar quote_char, gchar escape_char) {
@@ -260,36 +253,6 @@ quote_with_squotes(const gchar* value)
 	return quote_generic(value, '\'', '\'');
 }
 
-MDB_DEPRECATED(char*,
-mdb_get_coltype_string(MdbBackend *backend, int col_type))
-{
-	static int warn_deprecated = 0;
-	static char buf[16];
-
-	if (!warn_deprecated) {
-		warn_deprecated = 1;
-		fprintf(stderr, "mdb_get_coltype_string is deprecated. Use mdb_get_colbacktype_string.\n");
-	}
-	if (col_type > 0x10 ) {
-   		// return NULL;
-		snprintf(buf,sizeof(buf), "type %04x", col_type);
-		return buf;
-	} else
-		return backend->types_table[col_type].name;
-}
-
-MDB_DEPRECATED(int,
-mdb_coltype_takes_length(MdbBackend *backend, int col_type))
-{
-	static int warn_deprecated = 0;
-	if (!warn_deprecated) {
-		warn_deprecated = 1;
-		fprintf(stderr, "mdb_coltype_takes_length is deprecated. Use mdb_colbacktype_takes_length.\n");
-	}
-	return backend->types_table[col_type].needs_length;
-}
-
-
 const MdbBackendType*
 mdb_get_colbacktype(const MdbColumn *col) {
 	MdbBackend *backend = col->table->entry->mdb->default_backend;
@@ -299,8 +262,7 @@ mdb_get_colbacktype(const MdbColumn *col) {
 	if (col_type == MDB_LONGINT && col->is_long_auto && backend->type_autonum)
 		return backend->type_autonum;
 	if (col_type == MDB_DATETIME && backend->type_shortdate) {
-		const char *format = mdb_col_get_prop(col, "Format");
-		if (format && !strcmp(format, "Short Date"))
+		if (mdb_col_is_shortdate(col))
 			return backend->type_shortdate;
 	}
 	return &backend->types_table[col_type];
@@ -312,8 +274,8 @@ mdb_get_colbacktype_string(const MdbColumn *col)
 	const MdbBackendType *type = mdb_get_colbacktype(col);
 	if (!type) {
    		// return NULL;
-		static char buf[16];
-		snprintf(buf,sizeof(buf), "type %04x", col->col_type);
+		static __thread char buf[16];
+		snprintf(buf, sizeof(buf), "type %04x", col->col_type);
 		return buf;
 	}
 	return type->name;
@@ -326,108 +288,135 @@ mdb_colbacktype_takes_length(const MdbColumn *col)
 	return type->needs_length;
 }
 
-MDB_DEPRECATED(void,
-mdb_init_backends())
-{
-	fprintf(stderr, "mdb_init_backends() is DEPRECATED and does nothing. Stop calling it.\n");
-}
-
 /**
- * _mdb_init_backends
+ * mdb_init_backends
  *
  * Initializes the mdb_backends hash and loads the builtin backends.
  * Use mdb_remove_backends() to destroy this hash when done.
  */
-MDB_CONSTRUCTOR(_mdb_init_backends)
+void mdb_init_backends(MdbHandle *mdb)
 {
-	mdb_backends = g_hash_table_new(g_str_hash, g_str_equal);
+    if (mdb->backends) {
+        mdb_remove_backends(mdb);
+    }
+	mdb->backends = g_hash_table_new(g_str_hash, g_str_equal);
 
-	mdb_register_backend("access",
+	mdb_register_backend(mdb, "access",
 		MDB_SHEXP_DROPTABLE|MDB_SHEXP_CST_NOTNULL|MDB_SHEXP_DEFVALUES,
 		mdb_access_types, NULL, NULL,
 		"Date()", "Date()",
+		NULL,
+		NULL,
 		"-- That file uses encoding %s\n",
 		"DROP TABLE %s;\n",
 		NULL,
 		NULL,
 		NULL,
+		NULL,
+		NULL,
 		quote_schema_name_bracket_merge);
-	mdb_register_backend("sybase",
+	mdb_register_backend(mdb, "sybase",
 		MDB_SHEXP_DROPTABLE|MDB_SHEXP_CST_NOTNULL|MDB_SHEXP_CST_NOTEMPTY|MDB_SHEXP_COMMENTS|MDB_SHEXP_DEFVALUES,
 		mdb_sybase_types, &mdb_sybase_shortdate_type, NULL,
 		"getdate()", "getdate()",
+		NULL,
+		NULL,
 		"-- That file uses encoding %s\n",
 		"DROP TABLE %s;\n",
 		"ALTER TABLE %s ADD CHECK (%s <>'');\n",
 		"COMMENT ON COLUMN %s.%s IS %s;\n",
+		NULL,
 		"COMMENT ON TABLE %s IS %s;\n",
+		NULL,
 		quote_schema_name_dquote);
-	mdb_register_backend("oracle",
+	mdb_register_backend(mdb, "oracle",
 		MDB_SHEXP_DROPTABLE|MDB_SHEXP_CST_NOTNULL|MDB_SHEXP_COMMENTS|MDB_SHEXP_INDEXES|MDB_SHEXP_RELATIONS|MDB_SHEXP_DEFVALUES,
 		mdb_oracle_types, &mdb_oracle_shortdate_type, NULL,
 		"current_date", "sysdate",
+		NULL,
+		NULL,
 		"-- That file uses encoding %s\n",
 		"DROP TABLE %s;\n",
 		NULL,
 		"COMMENT ON COLUMN %s.%s IS %s;\n",
+		NULL,
 		"COMMENT ON TABLE %s IS %s;\n",
+		NULL,
 		quote_schema_name_dquote);
-	mdb_register_backend("postgres",
-		MDB_SHEXP_DROPTABLE|MDB_SHEXP_CST_NOTNULL|MDB_SHEXP_CST_NOTEMPTY|MDB_SHEXP_COMMENTS|MDB_SHEXP_INDEXES|MDB_SHEXP_RELATIONS|MDB_SHEXP_DEFVALUES,
+	mdb_register_backend(mdb, "postgres",
+		MDB_SHEXP_DROPTABLE|MDB_SHEXP_CST_NOTNULL|MDB_SHEXP_CST_NOTEMPTY|MDB_SHEXP_COMMENTS|MDB_SHEXP_INDEXES|MDB_SHEXP_RELATIONS|MDB_SHEXP_DEFVALUES|MDB_SHEXP_BULK_INSERT,
 		mdb_postgres_types, &mdb_postgres_shortdate_type, &mdb_postgres_serial_type,
 		"current_date", "now()",
+		"%Y-%m-%d %H:%M:%S",
+		"%Y-%m-%d",
 		"SET client_encoding = '%s';\n",
 		"DROP TABLE IF EXISTS %s;\n",
 		"ALTER TABLE %s ADD CHECK (%s <>'');\n",
 		"COMMENT ON COLUMN %s.%s IS %s;\n",
+		NULL,
 		"COMMENT ON TABLE %s IS %s;\n",
+		NULL,
 		quote_schema_name_dquote);
-	mdb_register_backend("mysql",
-		MDB_SHEXP_DROPTABLE|MDB_SHEXP_CST_NOTNULL|MDB_SHEXP_CST_NOTEMPTY|MDB_SHEXP_DEFVALUES,
+	mdb_register_backend(mdb, "mysql",
+		MDB_SHEXP_DROPTABLE|MDB_SHEXP_CST_NOTNULL|MDB_SHEXP_CST_NOTEMPTY|MDB_SHEXP_INDEXES|MDB_SHEXP_DEFVALUES|MDB_SHEXP_BULK_INSERT,
 		mdb_mysql_types, &mdb_mysql_shortdate_type, NULL,
 		"current_date", "now()",
+		"%Y-%m-%d %H:%M:%S",
+		"%Y-%m-%d",
 		"-- That file uses encoding %s\n",
 		"DROP TABLE IF EXISTS %s;\n",
 		"ALTER TABLE %s ADD CHECK (%s <>'');\n",
 		NULL,
+		"COMMENT %s",
 		NULL,
+		"COMMENT %s",
 		quote_schema_name_rquotes_merge);
-	mdb_register_backend("sqlite",
-		MDB_SHEXP_DROPTABLE|MDB_SHEXP_RELATIONS|MDB_SHEXP_DEFVALUES,
+	mdb_register_backend(mdb, "sqlite",
+		MDB_SHEXP_DROPTABLE|MDB_SHEXP_DEFVALUES|MDB_SHEXP_BULK_INSERT,
 		mdb_sqlite_types, NULL, NULL,
 		"date('now')", "date('now')",
+		"%Y-%m-%d %H:%M:%S",
+		"%Y-%m-%d",
 		"-- That file uses encoding %s\n",
 		"DROP TABLE IF EXISTS %s;\n",
 		NULL,
 		NULL,
 		NULL,
+		NULL,
+		NULL,
 		quote_schema_name_rquotes_merge);
-
-	atexit(_mdb_remove_backends);
 }
 
-void mdb_register_backend(char *backend_name, guint32 capabilities, MdbBackendType *backend_type, MdbBackendType *type_shortdate, MdbBackendType *type_autonum, const char *short_now, const char *long_now, const char *charset_statement, const char *drop_statement, const char *constaint_not_empty_statement, const char *column_comment_statement, const char *table_comment_statement, gchar* (*quote_schema_name)(const gchar*, const gchar*))
+void mdb_register_backend(MdbHandle *mdb, char *backend_name, guint32 capabilities,
+        const MdbBackendType *backend_type, const MdbBackendType *type_shortdate, const MdbBackendType *type_autonum,
+        const char *short_now, const char *long_now,
+        const char *date_fmt, const char *shortdate_fmt,
+        const char *charset_statement, const char *drop_statement,
+        const char *constaint_not_empty_statement,
+        const char *column_comment_statement,
+        const char *per_column_comment_statement, 
+        const char *table_comment_statement,
+        const char *per_table_comment_statement,
+        gchar* (*quote_schema_name)(const gchar*, const gchar*))
 {
-	MdbBackend *backend = (MdbBackend *) g_malloc0(sizeof(MdbBackend));
+	MdbBackend *backend = g_malloc0(sizeof(MdbBackend));
 	backend->capabilities = capabilities;
 	backend->types_table = backend_type;
 	backend->type_shortdate = type_shortdate;
 	backend->type_autonum = type_autonum;
 	backend->short_now = short_now;
 	backend->long_now = long_now;
+	backend->date_fmt = date_fmt;
+	backend->shortdate_fmt = shortdate_fmt;
 	backend->charset_statement = charset_statement;
 	backend->drop_statement = drop_statement;
 	backend->constaint_not_empty_statement = constaint_not_empty_statement;
 	backend->column_comment_statement = column_comment_statement;
+	backend->per_column_comment_statement = per_column_comment_statement;
 	backend->table_comment_statement = table_comment_statement;
+	backend->per_table_comment_statement = per_table_comment_statement;
 	backend->quote_schema_name  = quote_schema_name;
-	g_hash_table_insert(mdb_backends, backend_name, backend);
-}
-
-MDB_DEPRECATED(void,
-mdb_remove_backends())
-{
-	fprintf(stderr, "mdb_remove_backends() is DEPRECATED and does nothing. Stop calling it.\n");
+	g_hash_table_insert(mdb->backends, backend_name, backend);
 }
 
 /**
@@ -436,16 +425,15 @@ mdb_remove_backends())
  * Removes all entries from and destroys the mdb_backends hash.
  */
 void
-_mdb_remove_backends()
+mdb_remove_backends(MdbHandle *mdb)
 {
-	g_hash_table_foreach_remove(mdb_backends, mdb_drop_backend, NULL);
-	g_hash_table_destroy(mdb_backends);
+	g_hash_table_foreach(mdb->backends, mdb_drop_backend, NULL);
+	g_hash_table_destroy(mdb->backends);
 }
-static gboolean mdb_drop_backend(gpointer key, gpointer value, gpointer data)
+static void mdb_drop_backend(gpointer key, gpointer value, gpointer data)
 {
 	MdbBackend *backend = (MdbBackend *)value;
 	g_free (backend);
-	return TRUE;
 }
 
 /**
@@ -461,18 +449,68 @@ int mdb_set_default_backend(MdbHandle *mdb, const char *backend_name)
 {
 	MdbBackend *backend;
 
-	backend = (MdbBackend *) g_hash_table_lookup(mdb_backends, backend_name);
+    if (!mdb->backends) {
+        mdb_init_backends(mdb);
+    }
+	backend = (MdbBackend *) g_hash_table_lookup(mdb->backends, backend_name);
 	if (backend) {
 		mdb->default_backend = backend;
 		g_free(mdb->backend_name); // NULL is ok
 		mdb->backend_name = (char *) g_strdup(backend_name);
-		is_init = 0;
+		mdb->relationships_table = NULL;
+		if (backend->date_fmt) {
+			mdb_set_date_fmt(mdb, backend->date_fmt);
+		} else {
+			mdb_set_date_fmt(mdb, "%x %X");
+		}
+		if (backend->shortdate_fmt) {
+			mdb_set_shortdate_fmt(mdb, backend->shortdate_fmt);
+		} else {
+			mdb_set_shortdate_fmt(mdb, "%x");
+		}
 		return 1;
 	} else {
 		return 0;
 	}
 }
 
+
+/**
+ * Generates index name based on backend.
+ *
+ * You should free() the returned value once you are done with it.
+ *
+ * @param backend backend we are generating indexes for
+ * @param table table being processed
+ * @param idx index being processed
+ * @return the index name
+ */
+static char *
+mdb_get_index_name(int backend, MdbTableDef *table, MdbIndex *idx)
+{
+	char *index_name;
+
+	switch(backend){
+		case MDB_BACKEND_MYSQL:
+			// appending table name to index often makes it too long for mysql
+			if (idx->index_type==1)
+				// for mysql name of primary key is not used
+				index_name = g_strdup("_pkey");
+			else {
+				index_name = g_strdup(idx->name);
+			}
+			break;
+		default:
+			if (idx->index_type==1)
+				index_name = g_strconcat(table->name, "_pkey", NULL);
+			else {
+				index_name = g_strconcat(table->name, "_", idx->name, "_idx", NULL);
+			}
+	}
+
+	return index_name;
+}
+
 /**
  * mdb_print_indexes
  * @output: Where to print the sql
@@ -486,11 +524,18 @@ mdb_print_indexes(FILE* outfile, MdbTableDef *table, char *dbnamespace)
 	char* quoted_table_name;
 	char* index_name;
 	char* quoted_name;
+	int backend;
 	MdbHandle* mdb = table->entry->mdb;
 	MdbIndex *idx;
 	MdbColumn *col;
 
-	if (strcmp(mdb->backend_name, "postgres")) {
+	if (!strcmp(mdb->backend_name, "postgres")) {
+		backend = MDB_BACKEND_POSTGRES;
+	} else if (!strcmp(mdb->backend_name, "mysql")) {
+		backend = MDB_BACKEND_MYSQL;
+	} else if (!strcmp(mdb->backend_name, "oracle")) {
+		backend = MDB_BACKEND_ORACLE;
+	} else {
 		fprintf(outfile, "-- Indexes are not implemented for %s\n\n", mdb->backend_name);
 		return;
 	}
@@ -507,23 +552,47 @@ mdb_print_indexes(FILE* outfile, MdbTableDef *table, char *dbnamespace)
 		if (idx->index_type==2)
 			continue;
 
-		index_name = malloc(strlen(table->name)+strlen(idx->name)+5+1);
-		strcpy(index_name, table->name);
-		if (idx->index_type==1)
-			strcat(index_name, "_pkey");
-		else {
-			strcat(index_name, "_");
-			strcat(index_name, idx->name);
-			strcat(index_name, "_idx");
+		index_name = mdb_get_index_name(backend, table, idx);
+		switch (backend) {
+			case MDB_BACKEND_POSTGRES:
+				/* PostgreSQL index and constraint names are
+                                 * never namespaced in DDL (they are always
+                                 * created in same namespace as table), so
+                                 * omit namespace.
+                                 */
+				quoted_name = mdb->default_backend->quote_schema_name(NULL, index_name);
+				break;
+
+                         default:
+				quoted_name = mdb->default_backend->quote_schema_name(dbnamespace, index_name);
 		}
-		quoted_name = mdb->default_backend->quote_schema_name(dbnamespace, index_name);
+
 		if (idx->index_type==1) {
-			fprintf (outfile, "ALTER TABLE %s ADD CONSTRAINT %s PRIMARY KEY (", quoted_table_name, quoted_name);
+			switch (backend) {
+				case MDB_BACKEND_ORACLE:
+				case MDB_BACKEND_POSTGRES:
+					fprintf (outfile, "ALTER TABLE %s ADD CONSTRAINT %s PRIMARY KEY (", quoted_table_name, quoted_name);
+					break;
+				case MDB_BACKEND_MYSQL:
+					fprintf (outfile, "ALTER TABLE %s ADD PRIMARY KEY (", quoted_table_name);
+					break;
+			}
 		} else {
-			fprintf(outfile, "CREATE");
-			if (idx->flags & MDB_IDX_UNIQUE)
-				fprintf (outfile, " UNIQUE");
-			fprintf(outfile, " INDEX %s ON %s (", quoted_name, quoted_table_name);
+			switch (backend) {
+				case MDB_BACKEND_ORACLE:
+				case MDB_BACKEND_POSTGRES:
+					fprintf(outfile, "CREATE");
+					if (idx->flags & MDB_IDX_UNIQUE)
+						fprintf (outfile, " UNIQUE");
+					fprintf(outfile, " INDEX %s ON %s (", quoted_name, quoted_table_name);
+					break;
+				case MDB_BACKEND_MYSQL:
+					fprintf(outfile, "ALTER TABLE %s ADD", quoted_table_name);
+					if (idx->flags & MDB_IDX_UNIQUE)
+						fprintf (outfile, " UNIQUE");
+					fprintf(outfile, " INDEX %s (", quoted_name);
+					break;
+			}
 		}
 		g_free(quoted_name);
 		free(index_name);
@@ -544,7 +613,7 @@ mdb_print_indexes(FILE* outfile, MdbTableDef *table, char *dbnamespace)
 		fprintf (outfile, ");\n");
 	}
 	fputc ('\n', outfile);
-	
+
 	g_free(quoted_table_name);
 }
 
@@ -571,69 +640,51 @@ mdb_get_relationships(MdbHandle *mdb, const gchar *dbnamespace, const char* tabl
 {
 	unsigned int i;
 	gchar *text = NULL;  /* String to be returned */
-	static char *bound[5];  /* Bound values */
-	static MdbTableDef *table;  /* Relationships table */
-	int backend = 0;  /* Backends: 1=oracle, 2=postgres */
+	char **bound = mdb->relationships_values;  /* Bound values */
+	int backend = 0;
 	char *quoted_table_1, *quoted_column_1,
 	     *quoted_table_2, *quoted_column_2,
 	     *constraint_name, *quoted_constraint_name;
 	long grbit;
 
 	if (!strcmp(mdb->backend_name, "oracle")) {
-		backend = 1;
+		backend = MDB_BACKEND_ORACLE;
 	} else if (!strcmp(mdb->backend_name, "postgres")) {
-		backend = 2;
-	} else if (!strcmp(mdb->backend_name, "sqlite")) {
-		backend = 3;
-	} else {
-		if (is_init == 0) { /* the first time through */
-			is_init = 1;
-			return (char *) g_strconcat(
-				"-- relationships are not implemented for ",
-				mdb->backend_name, "\n", NULL);
-		} else { /* the second time through */
-			is_init = 0;
-			return NULL;
-		}
+		backend = MDB_BACKEND_POSTGRES;
+	} else if (!mdb->relationships_table) {
+		return NULL;
 	}
 
-	if (is_init == 0) {
-		table = mdb_read_table_by_name(mdb, "MSysRelationships", MDB_TABLE);
-		if ((!table) || (table->num_rows == 0)) {
+	if (!mdb->relationships_table) {
+		mdb->relationships_table = mdb_read_table_by_name(mdb, "MSysRelationships", MDB_TABLE);
+		if (!mdb->relationships_table || !mdb->relationships_table->num_rows) {
 			fprintf(stderr, "No MSysRelationships\n");
 			return NULL;
 		}
 
-		mdb_read_columns(table);
+		mdb_read_columns(mdb->relationships_table);
 		for (i=0;i<5;i++) {
-			bound[i] = (char *) g_malloc0(MDB_BIND_SIZE);
-		}
-		mdb_bind_column_by_name(table, "szColumn", bound[0], NULL);
-		mdb_bind_column_by_name(table, "szObject", bound[1], NULL);
-		mdb_bind_column_by_name(table, "szReferencedColumn", bound[2], NULL);
-		mdb_bind_column_by_name(table, "szReferencedObject", bound[3], NULL);
-		mdb_bind_column_by_name(table, "grbit", bound[4], NULL);
-		mdb_rewind_table(table);
-
-		is_init = 1;
-	}
-	else {
-		if (!table) {
-			fprintf(stderr, "table is NULL\n");
-		}
-	    if (table->cur_row >= table->num_rows) {  /* past the last row */
-			for (i=0;i<5;i++)
-				g_free(bound[i]);
-			is_init = 0;
-			return NULL;
+			bound[i] = g_malloc0(mdb->bind_size);
 		}
+		mdb_bind_column_by_name(mdb->relationships_table, "szColumn", bound[0], NULL);
+		mdb_bind_column_by_name(mdb->relationships_table, "szObject", bound[1], NULL);
+		mdb_bind_column_by_name(mdb->relationships_table, "szReferencedColumn", bound[2], NULL);
+		mdb_bind_column_by_name(mdb->relationships_table, "szReferencedObject", bound[3], NULL);
+		mdb_bind_column_by_name(mdb->relationships_table, "grbit", bound[4], NULL);
+		mdb_rewind_table(mdb->relationships_table);
 	}
+    if (mdb->relationships_table->cur_row >= mdb->relationships_table->num_rows) {  /* past the last row */
+        for (i=0;i<5;i++)
+            g_free(bound[i]);
+        mdb->relationships_table = NULL;
+        return NULL;
+    }
 
 	while (1) {
-		if (!mdb_fetch_row(table)) {
+		if (!mdb_fetch_row(mdb->relationships_table)) {
 			for (i=0;i<5;i++)
 				g_free(bound[i]);
-			is_init = 0;
+			mdb->relationships_table = NULL;
 			return NULL;
 		}
 		if (!tablename || !strcmp(bound[1], tablename))
@@ -641,12 +692,32 @@ mdb_get_relationships(MdbHandle *mdb, const gchar *dbnamespace, const char* tabl
 	}
 
 	quoted_table_1 = mdb->default_backend->quote_schema_name(dbnamespace, bound[1]);
-	quoted_column_1 = mdb->default_backend->quote_schema_name(dbnamespace, bound[0]);
 	quoted_table_2 = mdb->default_backend->quote_schema_name(dbnamespace, bound[3]);
-	quoted_column_2 = mdb->default_backend->quote_schema_name(dbnamespace, bound[2]);
 	grbit = atoi(bound[4]);
 	constraint_name = g_strconcat(bound[1], "_", bound[0], "_fk", NULL);
-	quoted_constraint_name = mdb->default_backend->quote_schema_name(dbnamespace, constraint_name);
+
+	switch (backend) {
+		case MDB_BACKEND_POSTGRES:
+			/* PostgreSQL index and constraint names are
+			 * never namespaced in DDL (they are always
+			 * created in same namespace as table), so
+			 * omit namespace.  Nor should column names
+                         * be namespaced.
+			 */
+			quoted_constraint_name = mdb->default_backend->quote_schema_name(NULL, constraint_name);
+			quoted_column_1 = mdb->default_backend->quote_schema_name(NULL, bound[0]);
+			quoted_column_2 = mdb->default_backend->quote_schema_name(NULL, bound[2]);
+			break;
+
+		default:
+			/* Other databases, namespace constraint and
+			 * column names.
+			 */
+			quoted_constraint_name = mdb->default_backend->quote_schema_name(dbnamespace, constraint_name);
+			quoted_column_1 = mdb->default_backend->quote_schema_name(dbnamespace, bound[0]);
+			quoted_column_2 = mdb->default_backend->quote_schema_name(dbnamespace, bound[2]);
+			break;
+	}
 	g_free(constraint_name);
 
 	if (grbit & 0x00000002) {
@@ -657,9 +728,17 @@ mdb_get_relationships(MdbHandle *mdb, const gchar *dbnamespace, const char* tabl
 			" does not enforce integrity.\n", NULL);
 	} else {
 		switch (backend) {
-		  case 1:  /* oracle */
-		  case 2:  /* postgres */
-		  case 3:  /* sqlite */
+		  case MDB_BACKEND_ORACLE:
+                        text = g_strconcat(
+                                "ALTER TABLE ", quoted_table_1,
+                                " ADD CONSTRAINT ", quoted_constraint_name,
+                                " FOREIGN KEY (", quoted_column_1, ")"
+                                " REFERENCES ", quoted_table_2, "(", quoted_column_2, ")",
+                                (grbit & 0x00001000) ? " ON DELETE CASCADE" : "",
+                                ";\n", NULL);
+
+                        break;
+		  case MDB_BACKEND_POSTGRES:
 			text = g_strconcat(
 				"ALTER TABLE ", quoted_table_1,
 				" ADD CONSTRAINT ", quoted_constraint_name,
@@ -667,6 +746,12 @@ mdb_get_relationships(MdbHandle *mdb, const gchar *dbnamespace, const char* tabl
 				" REFERENCES ", quoted_table_2, "(", quoted_column_2, ")",
 				(grbit & 0x00000100) ? " ON UPDATE CASCADE" : "",
 				(grbit & 0x00001000) ? " ON DELETE CASCADE" : "",
+				/* On some databases (eg PostgreSQL) we also want to set
+				 * the constraints to be optionally deferrable, to
+				 * facilitate out of order bulk loading.
+				 */
+				" DEFERRABLE",
+				" INITIALLY IMMEDIATE",
 				";\n", NULL);
 
 			break;
@@ -718,12 +803,25 @@ generate_table_schema(FILE *outfile, MdbCatalogEntry *entry, char *dbnamespace,
 		g_free(quoted_name);
 
 		if (mdb_colbacktype_takes_length(col)) {
-
 			/* more portable version from DW patch */
 			if (col->col_size == 0)
-	    			fputs(" (255)", outfile);
+				fputs(" (255)", outfile);
+			else if (col->col_scale != 0)
+				fprintf(outfile, " (%d, %d)", col->col_scale, col->col_prec);
 			else
-	    			fprintf(outfile, " (%d)", col->col_size);
+				fprintf(outfile, " (%d)", col->col_size);
+		}
+
+		if (mdb->default_backend->per_column_comment_statement && export_options & MDB_SHEXP_COMMENTS) {
+			prop_value = mdb_col_get_prop(col, "Description");
+			if (prop_value) {
+				char *comment = quote_with_squotes(prop_value);
+				fputs(" ", outfile);
+				fprintf(outfile,
+					mdb->default_backend->per_column_comment_statement,
+					comment);
+				free(comment);
+			}
 		}
 
 		if (export_options & MDB_SHEXP_CST_NOTNULL) {
@@ -759,8 +857,8 @@ generate_table_schema(FILE *outfile, MdbCatalogEntry *entry, char *dbnamespace,
 						fputs("TRUE", outfile);
 					else if (!strcmp(defval, "No"))
 						fputs("FALSE", outfile);
-					else if (!strcasecmp(defval, "date()")) {
-						if (!strcmp(mdb_col_get_prop(col, "Format"), "Short Date"))
+					else if (!g_ascii_strcasecmp(defval, "date()")) {
+						if (mdb_col_is_shortdate(col))
 							fputs(mdb->default_backend->short_now, outfile);
 						else
 							fputs(mdb->default_backend->long_now, outfile);
@@ -780,7 +878,17 @@ generate_table_schema(FILE *outfile, MdbCatalogEntry *entry, char *dbnamespace,
 			fputs("\n", outfile);
 	} /* for */
 
-	fputs(");\n", outfile);
+	fputs(")", outfile);
+	if (mdb->default_backend->per_table_comment_statement && export_options & MDB_SHEXP_COMMENTS) {
+		prop_value = mdb_table_get_prop(table, "Description");
+		if (prop_value) {
+			char *comment = quote_with_squotes(prop_value);
+			fputs(" ", outfile);
+			fprintf(outfile, mdb->default_backend->per_table_comment_statement, comment);
+			free(comment);
+		}
+	}
+	fputs(";\n", outfile);
 
 	/* Add the constraints on columns */
 	for (i = 0; i < table->num_cols; i++) {
@@ -799,7 +907,7 @@ generate_table_schema(FILE *outfile, MdbCatalogEntry *entry, char *dbnamespace,
 						quoted_table_name, quoted_name);
 		}
 
-		if (export_options & MDB_SHEXP_COMMENTS) {
+		if (mdb->default_backend->column_comment_statement && export_options & MDB_SHEXP_COMMENTS) {
 			prop_value = mdb_col_get_prop(col, "Description");
 			if (prop_value) {
 				char *comment = quote_with_squotes(prop_value);
@@ -813,8 +921,8 @@ generate_table_schema(FILE *outfile, MdbCatalogEntry *entry, char *dbnamespace,
 		g_free(quoted_name);
 	}
 
-	/* Add the comments on table */
-	if (export_options & MDB_SHEXP_COMMENTS) {
+	/* Add the constraints on table */
+	if (mdb->default_backend->table_comment_statement && export_options & MDB_SHEXP_COMMENTS) {
 		prop_value = mdb_table_get_prop(table, "Description");
 		if (prop_value) {
 			char *comment = quote_with_squotes(prop_value);
@@ -878,10 +986,85 @@ mdb_print_schema(MdbHandle *mdb, FILE *outfile, char *tabname, char *dbnamespace
 
 	if (export_options & MDB_SHEXP_RELATIONS) {
 		fputs ("-- CREATE Relationships ...\n", outfile);
-		while ((the_relation=mdb_get_relationships(mdb, dbnamespace, tabname)) != NULL) {
-			fputs(the_relation, outfile);
-			g_free(the_relation);
+		the_relation=mdb_get_relationships(mdb, dbnamespace, tabname);
+		if (!the_relation) {
+			fputs("-- relationships are not implemented for ", outfile);
+			fputs(mdb->backend_name, outfile);
+			fputs("\n", outfile);
+		} else {
+			do {
+				fputs(the_relation, outfile);
+				g_free(the_relation);
+			} while ((the_relation=mdb_get_relationships(mdb, dbnamespace, tabname)) != NULL);
 		}
 	}
 }
+
+#define MDB_BINEXPORT_MASK 0x0F
+#define is_binary_type(x) (x==MDB_OLE || x==MDB_BINARY || x==MDB_REPID)
+#define is_quote_type(x) (is_binary_type(x) || x==MDB_TEXT || x==MDB_MEMO || x==MDB_DATETIME)
+//#define DONT_ESCAPE_ESCAPE
+void
+mdb_print_col(FILE *outfile, gchar *col_val, int quote_text, int col_type, int bin_len,
+		char *quote_char, char *escape_char, int flags)
+/* quote_text: Don't quote if 0.
+ */
+{
+	size_t quote_len = strlen(quote_char); /* multibyte */
+
+	size_t orig_escape_len = escape_char ? strlen(escape_char) : 0;
+	int quoting = quote_text && is_quote_type(col_type);
+    int bin_mode = (flags & MDB_BINEXPORT_MASK);
+    int escape_cr_lf = !!(flags & MDB_EXPORT_ESCAPE_CONTROL_CHARS);
+
+	/* double the quote char if no escape char passed */
+	if (!escape_char)
+		escape_char = quote_char;
+
+	if (quoting)
+		fputs(quote_char, outfile);
+
+	while (1) {
+		if (is_binary_type(col_type)) {
+			if (bin_mode == MDB_BINEXPORT_STRIP)
+				break;
+			if (!bin_len--)
+				break;
+		} else /* use \0 sentry */
+			if (!*col_val)
+				break;
+
+		if (is_binary_type(col_type) && bin_mode == MDB_BINEXPORT_OCTAL) {
+			fprintf(outfile, "\\%03o", *(unsigned char*)col_val++);
+		} else if (is_binary_type(col_type) && bin_mode == MDB_BINEXPORT_HEXADECIMAL) {
+			fprintf(outfile, "%02X", *(unsigned char*)col_val++);
+		} else if (quoting && quote_len && !strncmp(col_val, quote_char, quote_len)) {
+			fprintf(outfile, "%s%s", escape_char, quote_char);
+			col_val += quote_len;
+#ifndef DONT_ESCAPE_ESCAPE
+		} else if (quoting && orig_escape_len && !strncmp(col_val, escape_char, orig_escape_len)) {
+			fprintf(outfile, "%s%s", escape_char, escape_char);
+			col_val += orig_escape_len;
 #endif
+		} else if (escape_cr_lf && is_quote_type(col_type) && *col_val=='\r') {
+			col_val++;
+			putc('\\', outfile);
+			putc('r', outfile);
+		} else if (escape_cr_lf && is_quote_type(col_type) && *col_val=='\n') {
+			col_val++;
+			putc('\\', outfile);
+			putc('n', outfile);
+		} else if (escape_cr_lf && is_quote_type(col_type) && *col_val=='\t') {
+			col_val++;
+			putc('\\', outfile);
+			putc('t', outfile);
+		} else if (escape_cr_lf && is_quote_type(col_type) && *col_val=='\\') {
+			col_val++;
+			putc('\\', outfile);
+			putc('\\', outfile);
+		} else
+			putc(*col_val++, outfile);
+	}
+	if (quoting)
+		fputs(quote_char, outfile);
+}
