@@ -18,14 +18,10 @@
 
 #include "mdbtools.h"
 
-#ifdef DMALLOC
-#include "dmalloc.h"
-#endif
-
-char *
+const char *
 mdb_get_objtype_string(int obj_type)
 {
-static char *type_name[] = {"Form",
+    static const char *type_name[] = {"Form",
 			"Table",
 			"Macro",
 			"System Table",
@@ -39,7 +35,7 @@ static char *type_name[] = {"Form",
 			"Database"
 		};
 
-	if (obj_type > 11) {
+	if (obj_type >= (int)(sizeof(type_name)/sizeof(type_name[0]))) {
 		return NULL;
 	} else {
 		return type_name[obj_type]; 
@@ -48,7 +44,7 @@ static char *type_name[] = {"Form",
 
 void mdb_free_catalog(MdbHandle *mdb)
 {
-	unsigned int i, j;
+	guint i, j;
 	MdbCatalogEntry *entry;
 
 	if ((!mdb) || (!mdb->catalog)) return;
@@ -57,8 +53,8 @@ void mdb_free_catalog(MdbHandle *mdb)
 		if (entry) {
 			if (entry->props) {
 				for (j=0; j<entry->props->len; j++)
-					mdb_free_props(g_array_index(entry->props, MdbProperties*, j));
-				g_array_free(entry->props, TRUE);
+					mdb_free_props(g_ptr_array_index(entry->props, j));
+				g_ptr_array_free(entry->props, TRUE);
 			}
 			g_free(entry);
 		}
@@ -71,13 +67,13 @@ GPtrArray *mdb_read_catalog (MdbHandle *mdb, int objtype)
 {
 	MdbCatalogEntry *entry, msysobj;
 	MdbTableDef *table;
-	char obj_id[256];
-	char obj_name[MDB_MAX_OBJ_NAME];
-	char obj_type[256];
-	char obj_flags[256];
-	char obj_props[MDB_BIND_SIZE];
+	char *obj_id = NULL;
+	char *obj_name = NULL;
+	char *obj_type = NULL;
+	char *obj_flags = NULL;
+	char *obj_props = NULL;
 	int type;
-	unsigned int i;
+	int i;
 	MdbColumn *col_props;
 	int kkd_size_ole;
 
@@ -86,25 +82,44 @@ GPtrArray *mdb_read_catalog (MdbHandle *mdb, int objtype)
 	mdb->catalog = g_ptr_array_new();
 	mdb->num_catalog = 0;
 
+	obj_id = malloc(mdb->bind_size);
+	obj_name = malloc(mdb->bind_size);
+	obj_type = malloc(mdb->bind_size);
+	obj_flags = malloc(mdb->bind_size);
+	obj_props = malloc(mdb->bind_size);
+
 	/* dummy up a catalog entry so we may read the table def */
 	memset(&msysobj, 0, sizeof(MdbCatalogEntry));
 	msysobj.mdb = mdb;
 	msysobj.object_type = MDB_TABLE;
 	msysobj.table_pg = 2;
-	strcpy(msysobj.object_name, "MSysObjects");
+	snprintf(msysobj.object_name, sizeof(msysobj.object_name), "%s", "MSysObjects");
 
 	/* mdb_table_dump(&msysobj); */
 
 	table = mdb_read_table(&msysobj);
-	if (!table) return NULL;
+    if (!table) {
+        fprintf(stderr, "Unable to read table %s\n", msysobj.object_name);
+        mdb_free_catalog(mdb);
+        goto cleanup;
+    }
 
 	mdb_read_columns(table);
 
-	mdb_bind_column_by_name(table, "Id", obj_id, NULL);
-	mdb_bind_column_by_name(table, "Name", obj_name, NULL);
-	mdb_bind_column_by_name(table, "Type", obj_type, NULL);
-	mdb_bind_column_by_name(table, "Flags", obj_flags, NULL);
-	i = mdb_bind_column_by_name(table, "LvProp", obj_props, &kkd_size_ole);
+    if (mdb_bind_column_by_name(table, "Id", obj_id, NULL) == -1 ||
+        mdb_bind_column_by_name(table, "Name", obj_name, NULL) == -1 ||
+        mdb_bind_column_by_name(table, "Type", obj_type, NULL) == -1 ||
+        mdb_bind_column_by_name(table, "Flags", obj_flags, NULL) == -1) {
+        fprintf(stderr, "Unable to bind columns from table %s (%d columns found)\n",
+                msysobj.object_name, table->num_cols);
+        mdb_free_catalog(mdb);
+        goto cleanup;
+    }
+    if ((i = mdb_bind_column_by_name(table, "LvProp", obj_props, &kkd_size_ole)) == -1) {
+        fprintf(stderr, "Unable to bind column %s from table %s\n", "LvProp", msysobj.object_name);
+        mdb_free_catalog(mdb);
+        goto cleanup;
+    }
 	col_props = g_ptr_array_index(table->columns, i-1);
 
 	mdb_rewind_table(table);
@@ -114,9 +129,9 @@ GPtrArray *mdb_read_catalog (MdbHandle *mdb, int objtype)
 		if (objtype==MDB_ANY || type == objtype) {
 			//fprintf(stderr, "obj_id: %10ld objtype: %-3d (0x%04x) obj_name: %s\n",
 			//	(atol(obj_id) & 0x00FFFFFF), type, type, obj_name);
-			entry = (MdbCatalogEntry *) g_malloc0(sizeof(MdbCatalogEntry));
+			entry = g_malloc0(sizeof(MdbCatalogEntry));
 			entry->mdb = mdb;
-			strcpy(entry->object_name, obj_name);
+			snprintf(entry->object_name, sizeof(entry->object_name), "%s", obj_name);
 			entry->object_type = (type & 0x7F);
 			entry->table_pg = atol(obj_id) & 0x00FFFFFF;
 			entry->flags = atol(obj_flags);
@@ -133,9 +148,17 @@ GPtrArray *mdb_read_catalog (MdbHandle *mdb, int objtype)
 	}
 	//mdb_dump_catalog(mdb, MDB_TABLE);
  
-	mdb_free_tabledef(table);
+cleanup:
+	if (table)
+		mdb_free_tabledef(table);
 
-	return mdb->catalog;
+	free(obj_id);
+	free(obj_name);
+	free(obj_type);
+	free(obj_flags);
+	free(obj_props);
+
+    return mdb->catalog;
 }
 
 
@@ -147,7 +170,7 @@ mdb_get_catalogentry_by_name(MdbHandle *mdb, const gchar* name)
 
 	for (i=0; i<mdb->num_catalog; i++) {
 		entry = g_ptr_array_index(mdb->catalog, i);
-		if (!strcasecmp(entry->object_name, name))
+		if (!g_ascii_strcasecmp(entry->object_name, name))
 			return entry;
 	}
 	return NULL;
