@@ -16,48 +16,88 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#include <time.h>
-#include <math.h>
 #include "mdbtools.h"
 
-#ifdef DMALLOC
-#include "dmalloc.h"
-#endif
+#define _XOPEN_SOURCE
+#include <time.h>
 
 #define OFFSET_MASK 0x1fff
+#define OLE_BUFFER_SIZE (MDB_BIND_SIZE*64)
 
 char *mdb_money_to_string(MdbHandle *mdb, int start);
 char *mdb_numeric_to_string(MdbHandle *mdb, int start, int prec, int scale);
 
 static int _mdb_attempt_bind(MdbHandle *mdb, 
 	MdbColumn *col, unsigned char isnull, int offset, int len);
-static char *mdb_date_to_string(MdbHandle *mdb, int start);
+static char *mdb_date_to_string(MdbHandle *mdb, const char *fmt, void *buf, int start);
 #ifdef MDB_COPY_OLE
 static size_t mdb_copy_ole(MdbHandle *mdb, void *dest, int start, int size);
 #endif
 
-static char date_fmt[64] = "%x %X";
+static const int noleap_cal[] = {0,31,59,90,120,151,181,212,243,273,304,334,365};
+static const int leap_cal[]   = {0,31,60,91,121,152,182,213,244,274,305,335,366};
 
-void mdb_set_date_fmt(const char *fmt)
-{
-		date_fmt[63] = 0; 
-		strncpy(date_fmt, fmt, 63);
+/* Some databases (eg PostgreSQL) do not understand integer 0/1 values
+ * as TRUE/FALSE, so provide a means to override the values used to be
+ * the SQL Standard TRUE/FALSE values.
+ */
+static const char boolean_false_number[] = "0";
+static const char boolean_true_number[]  = "1";
+
+static const char boolean_false_word[]   = "FALSE";
+static const char boolean_true_word[]    = "TRUE";
+
+void mdb_set_bind_size(MdbHandle *mdb, size_t bind_size) {
+    mdb->bind_size = bind_size;
 }
 
-void mdb_bind_column(MdbTableDef *table, int col_num, void *bind_ptr, int *len_ptr)
+void mdb_set_date_fmt(MdbHandle *mdb, const char *fmt)
 {
-	MdbColumn *col;
+    snprintf(mdb->date_fmt, sizeof(mdb->date_fmt), "%s", fmt);
+}
+
+void mdb_set_shortdate_fmt(MdbHandle *mdb, const char *fmt)
+{
+    snprintf(mdb->shortdate_fmt, sizeof(mdb->shortdate_fmt), "%s", fmt);
+}
+
+void mdb_set_boolean_fmt_numbers(MdbHandle *mdb)
+{
+	mdb->boolean_false_value = boolean_false_number;
+	mdb->boolean_true_value  = boolean_true_number;
+}
+
+void mdb_set_boolean_fmt_words(MdbHandle *mdb)
+{
+	mdb->boolean_false_value = boolean_false_word;
+	mdb->boolean_true_value  = boolean_true_word;
+}
+
+int mdb_bind_column(MdbTableDef *table, int col_num, void *bind_ptr, int *len_ptr)
+{
+	MdbColumn *col = NULL;
 
 	/* 
 	** the column arrary is 0 based, so decrement to get 1 based parameter 
 	*/
-	col=g_ptr_array_index(table->columns, col_num - 1);
-	
-	if (bind_ptr)
-		col->bind_ptr = bind_ptr;
-	if (len_ptr)
-		col->len_ptr = len_ptr;
+	col_num--;
+
+	if (col_num >= 0 && col_num < (int)table->num_cols) {
+		col=g_ptr_array_index(table->columns, col_num);
+
+		if (col) {
+			if (bind_ptr)
+				col->bind_ptr = bind_ptr;
+			if (len_ptr)
+				col->len_ptr = len_ptr;
+
+			return col_num + 1;
+		}
+	}
+
+	return -1;
 }
+
 int
 mdb_bind_column_by_name(MdbTableDef *table, gchar *col_name, void *bind_ptr, int *len_ptr)
 {
@@ -67,7 +107,7 @@ mdb_bind_column_by_name(MdbTableDef *table, gchar *col_name, void *bind_ptr, int
 	
 	for (i=0;i<table->num_cols;i++) {
 		col=g_ptr_array_index(table->columns,i);
-		if (!strcasecmp(col->name,col_name)) {
+		if (!g_ascii_strcasecmp(col->name,col_name)) {
 			col_num = i + 1;
 			if (bind_ptr)
 				col->bind_ptr = bind_ptr;
@@ -88,20 +128,22 @@ mdb_bind_column_by_name(MdbTableDef *table, gchar *col_name, void *bind_ptr, int
  * @off: Pointer for returning an offset to the row
  * @len: Pointer for returning the length of the row
  * 
- * Returns: 0 on success.  1 on failure.
+ * Returns: 0 on success. -1 on failure.
  */
 int mdb_find_pg_row(MdbHandle *mdb, int pg_row, void **buf, int *off, size_t *len)
 {
 	unsigned int pg = pg_row >> 8;
 	unsigned int row = pg_row & 0xff;
+    int result = 0;
 
 	if (mdb_read_alt_pg(mdb, pg) != mdb->fmt->pg_size)
-		return 1;
+		return -1;
 	mdb_swap_pgbuf(mdb);
-	mdb_find_row(mdb, row, off, len);
+	result = mdb_find_row(mdb, row, off, len);
 	mdb_swap_pgbuf(mdb);
+    *off &= OFFSET_MASK;
 	*buf = mdb->alt_pg_buf;
-	return 0;
+	return result;
 }
 
 int mdb_find_row(MdbHandle *mdb, int row, int *start, size_t *len)
@@ -115,6 +157,12 @@ int mdb_find_row(MdbHandle *mdb, int row, int *start, size_t *len)
 	next_start = (row == 0) ? mdb->fmt->pg_size :
 		mdb_get_int16(mdb->pg_buf, rco + row*2) & OFFSET_MASK;
 	*len = next_start - (*start & OFFSET_MASK);
+
+	if ((*start & OFFSET_MASK) >= mdb->fmt->pg_size ||
+			(*start & OFFSET_MASK) > next_start ||
+			next_start > mdb->fmt->pg_size)
+		return -1;
+
 	return 0;
 }
 
@@ -168,10 +216,11 @@ mdb_xfer_bound_bool(MdbHandle *mdb, MdbColumn *col, int value)
 {
 	col->cur_value_len = value;
 	if (col->bind_ptr) {
-		strcpy(col->bind_ptr, value ? "0" : "1");
+		strcpy(col->bind_ptr,
+                       value ? mdb->boolean_false_value : mdb->boolean_true_value);
 	}
 	if (col->len_ptr) {
-		*col->len_ptr = 1;
+		*col->len_ptr = strlen(col->bind_ptr);
 	}
 
 	return 1;
@@ -223,11 +272,17 @@ int ret;
 			//fprintf(stdout,"len %d size %d\n",len, col->col_size);
 			char *str;
 			if (col->col_type == MDB_NUMERIC) {
-				str = mdb_numeric_to_string(mdb, start, col->col_prec, col->col_scale);
+				str = mdb_numeric_to_string(mdb, start, col->col_scale, col->col_prec);
+			} else if (col->col_type == MDB_DATETIME) {
+				if (mdb_col_is_shortdate(col)) {
+					str = mdb_date_to_string(mdb, mdb->shortdate_fmt, mdb->pg_buf, start);
+				} else {
+					str = mdb_date_to_string(mdb, mdb->date_fmt, mdb->pg_buf, start);
+				}
 			} else {
 				str = mdb_col_to_string(mdb, mdb->pg_buf, start, col->col_type, len);
 			}
-			strcpy(col->bind_ptr, str);
+			snprintf(col->bind_ptr, mdb->bind_size, "%s", str);
 			g_free(str);
 		}
 		ret = strlen(col->bind_ptr);
@@ -244,16 +299,20 @@ int mdb_read_row(MdbTableDef *table, unsigned int row)
 	MdbColumn *col;
 	unsigned int i;
 	int row_start;
-	size_t row_size;
+	size_t row_size = 0;
 	int delflag, lookupflag;
-	MdbField fields[256];
+	MdbField *fields;
 	int num_fields;
 
-	if (table->num_rows == 0) 
+	if (table->num_cols == 0)
 		return 0;
 
-	if (mdb_find_row(mdb, row, &row_start, &row_size)) {
-		fprintf(stderr, "warning: mdb_find_row failed.");
+	if (mdb_find_row(mdb, row, &row_start, &row_size) == -1 || row_size == 0) {
+		/* Emitting a warning here isn't especially helpful. The row metadata
+		 * could be bogus for a number of reasons, so just skip to the next one
+		 * without comment. */
+		// fprintf(stderr, "warning: mdb_find_row failed.\n");
+		// fprintf(stderr, "warning: row_size = 0.\n");
 		return 0;
 	}
 
@@ -272,9 +331,13 @@ int mdb_read_row(MdbTableDef *table, unsigned int row)
 		return 0;
 	}
 
-	num_fields = mdb_crack_row(table, row_start, row_start + row_size - 1,
-		fields);
-	if (!mdb_test_sargs(table, fields, num_fields)) return 0;
+	fields = malloc(sizeof(MdbField) * table->num_cols);
+
+	num_fields = mdb_crack_row(table, row_start, row_size, fields);
+	if (num_fields < 0 || !mdb_test_sargs(table, fields, num_fields)) {
+		free(fields);
+		return 0;
+	}
 	
 #if MDB_DEBUG
 	fprintf(stdout,"sarg test passed row %d \n", row);
@@ -292,6 +355,8 @@ int mdb_read_row(MdbTableDef *table, unsigned int row)
 			fields[i].start, fields[i].siz);
 	}
 
+	free(fields);
+
 	return 1;
 }
 static int _mdb_attempt_bind(MdbHandle *mdb, 
@@ -337,7 +402,7 @@ int mdb_read_next_dpg(MdbTableDef *table)
 		}
 
 		table->cur_phys_pg = next_pg;
-		if (mdb->pg_buf[0]==MDB_PAGE_DATA && mdb_get_int32(mdb->pg_buf, 4)==entry->table_pg)
+		if (mdb->pg_buf[0]==MDB_PAGE_DATA && mdb_get_int32(mdb->pg_buf, 4)==(long)entry->table_pg)
 			return table->cur_phys_pg;
 
 		/* On rare occasion, mdb_map_find_next will return a wrong page */
@@ -352,7 +417,7 @@ int mdb_read_next_dpg(MdbTableDef *table)
 	do {
 		if (!mdb_read_pg(mdb, table->cur_phys_pg++))
 			return 0;
-	} while (mdb->pg_buf[0]!=MDB_PAGE_DATA || mdb_get_int32(mdb->pg_buf, 4)!=entry->table_pg);
+	} while (mdb->pg_buf[0]!=MDB_PAGE_DATA || mdb_get_int32(mdb->pg_buf, 4)!=(long)entry->table_pg);
 	/* fprintf(stderr,"returning new page %ld\n", table->cur_phys_pg); */
 	return table->cur_phys_pg;
 }
@@ -373,9 +438,6 @@ mdb_fetch_row(MdbTableDef *table)
 	int rc;
 	guint32 pg;
 
-	if (table->num_rows==0)
-		return 0;
-
 	/* initialize */
 	if (!table->cur_pg_num) {
 		table->cur_pg_num=1;
@@ -392,8 +454,7 @@ mdb_fetch_row(MdbTableDef *table)
 				fmt->row_count_offset);
 			if (table->cur_row >= rows) {
 				table->cur_row = 0;
-				table->cur_pg_num++;
-				if (table->cur_pg_num > pages->len)
+				if (++table->cur_pg_num > (unsigned int)pages->len)
 					return 0;
 			}
 			memcpy(mdb->pg_buf,
@@ -429,21 +490,30 @@ mdb_fetch_row(MdbTableDef *table)
 void mdb_data_dump(MdbTableDef *table)
 {
 	unsigned int i;
-	char *bound_values[MDB_MAX_COLS]; 
+	int ret;
+	char **bound_values = calloc(table->num_cols, sizeof(char *));
 
 	for (i=0;i<table->num_cols;i++) {
-		bound_values[i] = (char *) g_malloc(256);
-		mdb_bind_column(table, i+1, bound_values[i], NULL);
+		bound_values[i] = g_malloc(MDB_BIND_SIZE);
+		ret = mdb_bind_column(table, i+1, bound_values[i], NULL);
+		if (ret == -1) {
+			fprintf(stderr, "error binding column %d\n", i+1);
+			g_free(bound_values[i]);
+			bound_values[i] = NULL;
+		}
 	}
 	mdb_rewind_table(table);
 	while (mdb_fetch_row(table)) {
 		for (i=0;i<table->num_cols;i++) {
-			fprintf(stdout, "column %d is %s\n", i+1, bound_values[i]);
+			if (bound_values[i]) {
+				fprintf(stdout, "column %d is %s\n", i+1, bound_values[i]);
+			}
 		}
 	}
 	for (i=0;i<table->num_cols;i++) {
 		g_free(bound_values[i]);
 	}
+	free(bound_values);
 }
 
 int mdb_is_fixed_col(MdbColumn *col)
@@ -494,16 +564,18 @@ mdb_ole_read_next(MdbHandle *mdb, MdbColumn *col, void *ole_ptr)
 		&buf, &row_start, &len)) {
 		return 0;
 	}
+	if (len < 4)
+		return 0;
 	mdb_debug(MDB_DEBUG_OLE,"start %d len %d", row_start, len);
 
 	if (col->bind_ptr)
-		memcpy(col->bind_ptr, buf + row_start + 4, len - 4);
+		memcpy(col->bind_ptr, (char*)buf + row_start + 4, len - 4);
 	col->cur_blob_pg_row = mdb_get_int32(buf, row_start);
 
 	return len - 4;
 }
 size_t 
-mdb_ole_read(MdbHandle *mdb, MdbColumn *col, void *ole_ptr, int chunk_size)
+mdb_ole_read(MdbHandle *mdb, MdbColumn *col, void *ole_ptr, size_t chunk_size)
 {
 	guint32 ole_len;
 	void *buf;
@@ -519,16 +591,12 @@ mdb_ole_read(MdbHandle *mdb, MdbColumn *col, void *ole_ptr, int chunk_size)
 	if (ole_len & 0x80000000) {
 		/* inline ole field, if we can satisfy it, then do it */
 		len = col->cur_value_len - MDB_MEMO_OVERHEAD;
-		if (chunk_size >= len) {
-			if (col->bind_ptr) 
-				memcpy(col->bind_ptr, 
-					&mdb->pg_buf[col->cur_value_start + 
-						MDB_MEMO_OVERHEAD],
-					len);
-			return len;
-		} else {
+		if (chunk_size < len)
 			return 0;
-		}
+		if (col->bind_ptr)
+			memcpy(col->bind_ptr, &mdb->pg_buf[col->cur_value_start +
+					MDB_MEMO_OVERHEAD], len);
+		return len;
 	} else if (ole_len & 0x40000000) {
 		col->cur_blob_pg_row = mdb_get_int32(ole_ptr, 4);
 		mdb_debug(MDB_DEBUG_OLE,"ole row = %d ole pg = %ld",
@@ -542,25 +610,25 @@ mdb_ole_read(MdbHandle *mdb, MdbColumn *col, void *ole_ptr, int chunk_size)
 		mdb_debug(MDB_DEBUG_OLE,"start %d len %d", row_start, len);
 
 		if (col->bind_ptr) {
-			memcpy(col->bind_ptr, buf + row_start, len);
+			memcpy(col->bind_ptr, (char*)buf + row_start, len);
 			if (mdb_get_option(MDB_DEBUG_OLE))
 				mdb_buffer_dump(col->bind_ptr, 0, 16);
 		}
 		return len;
-	} else if ((ole_len & 0xff000000) == 0) {
+	} else if ((ole_len & 0xf0000000) == 0) {
 		col->cur_blob_pg_row = mdb_get_int32(ole_ptr, 4);
 		mdb_debug(MDB_DEBUG_OLE,"ole row = %d ole pg = %ld",
 			col->cur_blob_pg_row & 0xff,
 			col->cur_blob_pg_row >> 8);
 
 		if (mdb_find_pg_row(mdb, col->cur_blob_pg_row,
-			&buf, &row_start, &len)) {
+			&buf, &row_start, &len) || len < 4) {
 			return 0;
 		}
 		mdb_debug(MDB_DEBUG_OLE,"start %d len %d", row_start, len);
 
 		if (col->bind_ptr) 
-			memcpy(col->bind_ptr, buf + row_start + 4, len - 4);
+			memcpy(col->bind_ptr, (char*)buf + row_start + 4, len - 4);
 		col->cur_blob_pg_row = mdb_get_int32(buf, row_start);
 		mdb_debug(MDB_DEBUG_OLE, "next pg_row %d", col->cur_blob_pg_row);
 
@@ -574,24 +642,24 @@ mdb_ole_read(MdbHandle *mdb, MdbColumn *col, void *ole_ptr, int chunk_size)
  * mdb_ole_read_full calls mdb_ole_read then loop over mdb_ole_read_next as much as necessary.
  * returns the result in a big buffer.
  * The call must free it.
- * Note that this function is not indempotent: It may be called only once per column after each bind.
+ * Note that this function is not idempotent: It may be called only once per column after each bind.
  */
 void*
 mdb_ole_read_full(MdbHandle *mdb, MdbColumn *col, size_t *size)
 {
 	char ole_ptr[MDB_MEMO_OVERHEAD];
-	char *result = malloc(MDB_BIND_SIZE);
-	size_t result_buffer_size = MDB_BIND_SIZE;
+	char *result = malloc(OLE_BUFFER_SIZE);
+	size_t result_buffer_size = OLE_BUFFER_SIZE;
 	size_t len, pos;
 
 	memcpy(ole_ptr, col->bind_ptr, MDB_MEMO_OVERHEAD);
 
-	len = mdb_ole_read(mdb, col, ole_ptr, MDB_BIND_SIZE);
+	len = mdb_ole_read(mdb, col, ole_ptr, OLE_BUFFER_SIZE);
 	memcpy(result, col->bind_ptr, len);
 	pos = len;
 	while ((len = mdb_ole_read_next(mdb, col, ole_ptr))) {
 		if (pos+len >= result_buffer_size) {
-			result_buffer_size += MDB_BIND_SIZE;
+			result_buffer_size += OLE_BUFFER_SIZE;
 			result = realloc(result, result_buffer_size);
 		}
 		memcpy(result + pos, col->bind_ptr, len);
@@ -646,7 +714,7 @@ static size_t mdb_copy_ole(MdbHandle *mdb, void *dest, int start, int size)
 			mdb_debug(MDB_DEBUG_OLE,"Reading LVAL page %06x",
 				pg_row >> 8);
 
-			if (mdb_find_pg_row(mdb,pg_row,&buf,&row_start,&len)) {
+			if (mdb_find_pg_row(mdb,pg_row,&buf,&row_start,&len) || len < 4) {
 				return 0;
 			}
 
@@ -673,7 +741,7 @@ static char *mdb_memo_to_string(MdbHandle *mdb, int start, int size)
 	gint32 row_start, pg_row;
 	size_t len;
 	void *buf, *pg_buf = mdb->pg_buf;
-	char *text = (char *) g_malloc(MDB_BIND_SIZE);
+	char *text = g_malloc(mdb->bind_size);
 
 	if (size<MDB_MEMO_OVERHEAD) {
 		strcpy(text, "");
@@ -692,8 +760,8 @@ static char *mdb_memo_to_string(MdbHandle *mdb, int start, int size)
 
 	if (memo_len & 0x80000000) {
 		/* inline memo field */
-		mdb_unicode2ascii(mdb, pg_buf + start + MDB_MEMO_OVERHEAD,
-			size - MDB_MEMO_OVERHEAD, text, MDB_BIND_SIZE);
+		mdb_unicode2ascii(mdb, (char*)pg_buf + start + MDB_MEMO_OVERHEAD,
+			size - MDB_MEMO_OVERHEAD, text, mdb->bind_size);
 		return text;
 	} else if (memo_len & 0x40000000) {
 		/* single-page memo field */
@@ -710,14 +778,14 @@ static char *mdb_memo_to_string(MdbHandle *mdb, int start, int size)
 			pg_row & 0xff, row_start, len);
 		mdb_buffer_dump(buf, row_start, len);
 #endif
-		mdb_unicode2ascii(mdb, buf + row_start, len, text, MDB_BIND_SIZE);
+		mdb_unicode2ascii(mdb, (char*)buf + row_start, len, text, mdb->bind_size);
 		return text;
 	} else if ((memo_len & 0xff000000) == 0) { // assume all flags in MSB
 		/* multi-page memo field */
 		guint32 tmpoff = 0;
 		char *tmp;
 
-		tmp = (char *) g_malloc(memo_len);
+		tmp = g_malloc(memo_len);
 		pg_row = mdb_get_int32(pg_buf, start+4);
 		do {
 #if MDB_DEBUG
@@ -732,16 +800,20 @@ static char *mdb_memo_to_string(MdbHandle *mdb, int start, int size)
 			printf("row num %d start %d len %d\n",
 				pg_row & 0xff, row_start, len);
 #endif
-			if (tmpoff + len - 4 > memo_len) {
+			if (tmpoff + len - 4 > memo_len)
 				break;
-			}
-			memcpy(tmp + tmpoff, buf + row_start + 4, len - 4);
+
+			/* Stop processing on zero length multiple page memo fields */
+			if (len < 4)
+				break;
+
+			memcpy(tmp + tmpoff, (char*)buf + row_start + 4, len - 4);
 			tmpoff += len - 4;
 		} while (( pg_row = mdb_get_int32(buf, row_start) ));
 		if (tmpoff < memo_len) {
 			fprintf(stderr, "Warning: incorrect memo length\n");
 		}
-		mdb_unicode2ascii(mdb, tmp, tmpoff, text, MDB_BIND_SIZE);
+		mdb_unicode2ascii(mdb, tmp, tmpoff, text, mdb->bind_size);
 		g_free(tmp);
 		return text;
 	} else {
@@ -774,91 +846,99 @@ static int trim_trailing_zeros(char * buff)
 }
 #endif
 
+
 /* Date/Time is stored as a double, where the whole
    part is the days from 12/30/1899 and the fractional
    part is the fractional part of one day. */
 
+void
+mdb_tm_to_date(struct tm *t, double *td)
+{
+	short yr = t->tm_year + 1900;
+	char leap = ((yr & 3) == 0) && ((yr % 100) != 0 || (yr % 400) == 0);
+	const int *cal = leap ? leap_cal : noleap_cal;
+	long int time = (yr*365+(yr/4)-(yr/100)+(yr/400)+cal[t->tm_mon]+t->tm_mday)-693959;
+
+	*td = (((long)t->tm_hour * 3600)+((long)t->tm_min * 60)+((long)t->tm_sec)) / 86400.0;
+	if (time>=0) *td+=time; else *td=time-*td;
+}
+
 void
 mdb_date_to_tm(double td, struct tm *t)
 {
-	long int day, time;
-	int yr, q;
-	int *cal;
-	int noleap_cal[] = {0,31,59,90,120,151,181,212,243,273,304,334,365};
-	int leap_cal[]   = {0,31,60,91,121,152,182,213,244,274,305,335,366};
+	long day, time;
+	long yr, q;
+	const int *cal;
 
-	day = (long int)(td);
-	time = (long int)(fabs(td - day) * 86400.0 + 0.5);
+	if (td < 0.0 || td > 1e6) // About 2700 AD
+		return;
+
+	yr = 1;
+	day = (long)(td);
+	time = (long)((td - day) * 86400.0 + 0.5);
 	t->tm_hour = time / 3600;
 	t->tm_min = (time / 60) % 60;
 	t->tm_sec = time % 60;
-	t->tm_year = 1 - 1900;
 
 	day += 693593; /* Days from 1/1/1 to 12/31/1899 */
 	t->tm_wday = (day+1) % 7;
 
 	q = day / 146097;  /* 146097 days in 400 years */
-	t->tm_year += 400 * q;
+	yr += 400 * q;
 	day -= q * 146097;
 
 	q = day / 36524;  /* 36524 days in 100 years */
 	if (q > 3) q = 3;
-	t->tm_year += 100 * q;
+	yr += 100 * q;
 	day -= q * 36524;
 
 	q = day / 1461;  /* 1461 days in 4 years */
-	t->tm_year += 4 * q;
+	yr += 4 * q;
 	day -= q * 1461;
 
 	q = day / 365;  /* 365 days in 1 year */
 	if (q > 3) q = 3;
-	t->tm_year += q;
+	yr += q;
 	day -= q * 365;
 
-	yr = t->tm_year + 1900;
 	cal = ((yr)%4==0 && ((yr)%100!=0 || (yr)%400==0)) ?
 		leap_cal : noleap_cal;
 	for (t->tm_mon=0; t->tm_mon<12; t->tm_mon++) {
 		if (day < cal[t->tm_mon+1]) break;
 	}
+	t->tm_year = yr - 1900;
 	t->tm_mday = day - cal[t->tm_mon] + 1;
 	t->tm_yday = day;
 	t->tm_isdst = -1;
 }
 
 static char *
-mdb_date_to_string(MdbHandle *mdb, int start)
+mdb_date_to_string(MdbHandle *mdb, const char *fmt, void *buf, int start)
 {
-	struct tm t;
-	char *text = (char *) g_malloc(MDB_BIND_SIZE);
-	double td = mdb_get_double(mdb->pg_buf, start);
+	struct tm t = { 0 };
+	char *text = g_malloc(mdb->bind_size);
+	double td = mdb_get_double(buf, start);
 
 	mdb_date_to_tm(td, &t);
 
-	strftime(text, MDB_BIND_SIZE, date_fmt, &t);
+	strftime(text, mdb->bind_size, mdb->date_fmt, &t);
 
 	return text;
 }
 
-static char *
-mdb_uuid_to_string(MdbHandle *mdb, int start)
+char *mdb_uuid_to_string(const void *buf, int pos)
 {
-	char *text = NULL;
-  unsigned short uuid1, uuid2, uuid3, uuid4, uuid5, uuid6, uuid7, uuid8;
+	const unsigned char *kkd = (const unsigned char *)buf;
+	return g_strdup_printf("{%02X%02X%02X%02X" "-" "%02X%02X" "-" "%02X%02X"
+			"-" "%02X%02X" "%02X%02X%02X%02X%02X%02X}",
+			kkd[pos+3], kkd[pos+2], kkd[pos+1], kkd[pos], // little-endian
+			kkd[pos+5], kkd[pos+4], // little-endian
+			kkd[pos+7], kkd[pos+6], // little-endian
+			kkd[pos+8], kkd[pos+9], // big-endian
 
-  uuid1 = mdb_get_int16(mdb->pg_buf, start);
-  uuid2 = mdb_get_int16(mdb->pg_buf, start + 2);
-  uuid3 = mdb_get_int16(mdb->pg_buf, start + 4);
-  uuid4 = mdb_get_int16(mdb->pg_buf, start + 6);
-  uuid5 = mdb_get_int16(mdb->pg_buf, start + 8);
-  uuid6 = mdb_get_int16(mdb->pg_buf, start + 10);
-  uuid7 = mdb_get_int16(mdb->pg_buf, start + 12);
-  uuid8 = mdb_get_int16(mdb->pg_buf, start + 14);
-
-  text = g_strdup_printf("{%04x%04x-%04x-%04x-%04x-%04x%04x%04x}",
-    uuid1, uuid2, uuid3, uuid4, uuid5, uuid6, uuid7, uuid8);
-
-	return text;
+			kkd[pos+10], kkd[pos+11],
+			kkd[pos+12], kkd[pos+13],
+			kkd[pos+14], kkd[pos+15]); // big-endian
 }
 
 #if 0
@@ -917,31 +997,32 @@ char *mdb_col_to_string(MdbHandle *mdb, void *buf, int start, int datatype, int
 		break;
 		case MDB_FLOAT:
 			tf = mdb_get_single(buf, start);
-			text = g_strdup_printf("%.8e", tf);
+			text = g_strdup_printf("%.8g", tf);
 		break;
 		case MDB_DOUBLE:
 			td = mdb_get_double(buf, start);
-			text = g_strdup_printf("%.16e", td);
+			text = g_strdup_printf("%.16g", td);
 		break;
 		case MDB_BINARY:
 			if (size<0) {
 				text = g_strdup("");
 			} else {
-				text = g_malloc(size);
-				memcpy((char*)buf+start, text, size);
+				text = g_malloc(size+1);
+				memcpy(text, (char*)buf+start, size);
+				text[size] = '\0';
 			}
 		break;
 		case MDB_TEXT:
 			if (size<0) {
 				text = g_strdup("");
 			} else {
-				text = (char *) g_malloc(MDB_BIND_SIZE);
-				mdb_unicode2ascii(mdb, buf + start,
-					size, text, MDB_BIND_SIZE);
+				text = g_malloc(mdb->bind_size);
+				mdb_unicode2ascii(mdb, (char*)buf + start,
+					size, text, mdb->bind_size);
 			}
 		break;
 		case MDB_DATETIME:
-			text = mdb_date_to_string(mdb, start);
+			text = mdb_date_to_string(mdb, mdb->date_fmt, buf, start);
 		break;
 		case MDB_MEMO:
 			text = mdb_memo_to_string(mdb, start, size);
@@ -951,7 +1032,7 @@ char *mdb_col_to_string(MdbHandle *mdb, void *buf, int start, int datatype, int
 		case MDB_NUMERIC:
 		break;
 		case MDB_REPID:
-		  text = mdb_uuid_to_string(mdb, start);
+		  text = mdb_uuid_to_string(mdb->pg_buf, start);
 		break;
 		default:
 			text = g_strdup("");
