@@ -17,39 +17,48 @@
  */
 
 #include <inttypes.h>
+#include <stddef.h>
 #include "mdbtools.h"
 
-#ifdef DMALLOC
-#include "dmalloc.h"
-#endif
-
-/*
-typedef struct {
-	int		pg_size;
-	guint16		row_count_offset; 
-	guint16		tab_num_rows_offset;
-	guint16		tab_num_cols_offset;
-	guint16		tab_num_idxs_offset;
-	guint16		tab_num_ridxs_offset;
-	guint16		tab_usage_map_offset;
-	guint16		tab_first_dpg_offset;
-	guint16		tab_cols_start_offset;
-	guint16		tab_ridx_entry_size;
-	guint16		col_flags_offset;
-	guint16		col_size_offset;
-	guint16		col_num_offset;
-	guint16		tab_col_entry_size;
-	guint16         tab_free_map_offset;
-	guint16         tab_col_offset_var;
-	guint16         tab_col_offset_fixed;
-	guint16         tab_row_col_num_offset;
-} MdbFormatConstants; 
-*/
 MdbFormatConstants MdbJet4Constants = {
-	4096, 0x0c, 16, 45, 47, 51, 55, 56, 63, 12, 15, 23, 5, 25, 59, 7, 21, 9
+	.pg_size = 4096,
+    .row_count_offset = 0x0c,
+    .tab_num_rows_offset = 16,
+    .tab_num_cols_offset = 45,
+    .tab_num_idxs_offset = 47,
+    .tab_num_ridxs_offset = 51,
+    .tab_usage_map_offset = 55,
+    .tab_first_dpg_offset = 56,
+    .tab_cols_start_offset = 63,
+    .tab_ridx_entry_size = 12,
+    .col_flags_offset = 15,
+    .col_size_offset = 23,
+    .col_num_offset = 5,
+    .tab_col_entry_size = 25,
+    .tab_free_map_offset = 59,
+    .tab_col_offset_var = 7,
+    .tab_col_offset_fixed = 21,
+    .tab_row_col_num_offset = 9
 };
 MdbFormatConstants MdbJet3Constants = {
-	2048, 0x08, 12, 25, 27, 31, 35, 36, 43, 8, 13, 16, 1, 18, 39, 3, 14, 5
+	.pg_size = 2048,
+    .row_count_offset = 0x08,
+    .tab_num_rows_offset = 12,
+    .tab_num_cols_offset = 25,
+    .tab_num_idxs_offset = 27,
+    .tab_num_ridxs_offset = 31,
+    .tab_usage_map_offset = 35,
+    .tab_first_dpg_offset = 36,
+    .tab_cols_start_offset = 43,
+    .tab_ridx_entry_size = 8,
+    .col_flags_offset = 13,
+    .col_size_offset = 16,
+    .col_num_offset = 1,
+    .tab_col_entry_size = 18,
+    .tab_free_map_offset = 39,
+    .tab_col_offset_var = 3,
+    .tab_col_offset_fixed = 14,
+    .tab_row_col_num_offset = 5
 };
 
 typedef struct _RC4_KEY
@@ -159,6 +168,112 @@ static char *mdb_find_file(const char *file_name)
 	g_strfreev(dir);
 	return NULL;
 }
+
+/**
+ * mdb_handle_from_stream:
+ * @stream An open file stream
+ * @flags MDB_NOFLAGS for read-only, MDB_WRITABLE for read/write
+ *
+ * Allocates, initializes, and return an MDB handle from a file stream pointing
+ * to an MDB file.
+ *
+ * Return value: The handle on success, NULL on failure
+ */
+static MdbHandle *mdb_handle_from_stream(FILE *stream, MdbFileFlags flags) {
+	MdbHandle *mdb = g_malloc0(sizeof(MdbHandle));
+	mdb_set_default_backend(mdb, "access");
+    mdb_set_date_fmt(mdb, "%x %X");
+    mdb_set_shortdate_fmt(mdb, "%x");
+    mdb_set_bind_size(mdb, MDB_BIND_SIZE);
+    mdb_set_boolean_fmt_numbers(mdb);
+#ifdef HAVE_ICONV
+	mdb->iconv_in = (iconv_t)-1;
+	mdb->iconv_out = (iconv_t)-1;
+#endif
+	/* need something to bootstrap with, reassign after page 0 is read */
+	mdb->fmt = &MdbJet3Constants;
+	mdb->f = g_malloc0(sizeof(MdbFile));
+	mdb->f->refs = 1;
+	mdb->f->stream = stream;
+	if (flags & MDB_WRITABLE) {
+		mdb->f->writable = TRUE;
+    }
+
+	if (!mdb_read_pg(mdb, 0)) {
+		// fprintf(stderr,"Couldn't read first page.\n");
+		mdb_close(mdb);
+		return NULL;
+	}
+	if (mdb->pg_buf[0] != 0) {
+		mdb_close(mdb);
+		return NULL; 
+	}
+	mdb->f->jet_version = mdb_get_byte(mdb->pg_buf, 0x14);
+	switch(mdb->f->jet_version) {
+	case MDB_VER_JET3:
+		mdb->fmt = &MdbJet3Constants;
+		break;
+	case MDB_VER_JET4:
+	case MDB_VER_ACCDB_2007:
+	case MDB_VER_ACCDB_2010:
+	case MDB_VER_ACCDB_2013:
+	case MDB_VER_ACCDB_2016:
+		mdb->fmt = &MdbJet4Constants;
+		break;
+	default:
+		fprintf(stderr,"Unknown Jet version: %x\n", mdb->f->jet_version);
+		mdb_close(mdb);
+		return NULL; 
+	}
+
+    RC4_KEY rc4_key;
+    unsigned int tmp_key = 0x6b39dac7;
+    RC4_set_key(&rc4_key, 4, (unsigned char *)&tmp_key);
+    RC4(&rc4_key, mdb->f->jet_version == MDB_VER_JET3 ? 126 : 128, mdb->pg_buf + 0x18);
+
+	if (mdb->f->jet_version == MDB_VER_JET3) {
+		mdb->f->lang_id = mdb_get_int16(mdb->pg_buf, 0x3a);
+	} else {
+		mdb->f->lang_id = mdb_get_int16(mdb->pg_buf, 0x6e);
+	}
+	mdb->f->code_page = mdb_get_int16(mdb->pg_buf, 0x3c);
+	mdb->f->db_key = mdb_get_int32(mdb->pg_buf, 0x3e);
+    if (mdb->f->jet_version == MDB_VER_JET3) {
+        /* JET4 needs additional masking with the DB creation date, currently unsupported */
+        /* Bug - JET3 supports 20 byte passwords, this is currently just 14 bytes */
+        memcpy(mdb->f->db_passwd, mdb->pg_buf + 0x42, sizeof(mdb->f->db_passwd));
+    }
+    /* write is not supported for encrypted files yet */
+    mdb->f->writable = mdb->f->writable && !mdb->f->db_key;
+
+	mdb_iconv_init(mdb);
+
+	return mdb;
+}
+
+/**
+ * mdb_open_buffer:
+ * @buffer A memory buffer containing an MDB file
+ * @len Length of the buffer
+ *
+ * Opens an MDB file in memory and returns an MdbHandle to it.
+ *
+ * Return value: point to MdbHandle structure.
+ */
+MdbHandle *mdb_open_buffer(void *buffer, size_t len, MdbFileFlags flags) {
+    FILE *file = NULL;
+#ifdef HAVE_FMEMOPEN
+    file = fmemopen(buffer, len, (flags & MDB_WRITABLE) ? "r+" : "r");
+#else
+    fprintf(stderr, "mdb_open_buffer requires a platform with support for fmemopen(3)\n");
+#endif
+    if (file == NULL) {
+        fprintf(stderr, "Couldn't open memory buffer\n");
+        return NULL;
+    }
+    return mdb_handle_from_stream(file, flags);
+}
+
 /**
  * mdb_open:
  * @filename: path to MDB (database) file
@@ -172,109 +287,28 @@ static char *mdb_find_file(const char *file_name)
  **/
 MdbHandle *mdb_open(const char *filename, MdbFileFlags flags)
 {
-	MdbHandle *mdb;
-	int key[] = {0x86, 0xfb, 0xec, 0x37, 0x5d, 0x44, 0x9c, 0xfa, 0xc6, 0x5e, 0x28, 0xe6, 0x13, 0xb6};
-	int j, pos;
-	int open_flags;
+    FILE *file;
 
-	mdb = (MdbHandle *) g_malloc0(sizeof(MdbHandle));
-	mdb_set_default_backend(mdb, "access");
-#ifdef HAVE_ICONV
-	mdb->iconv_in = (iconv_t)-1;
-	mdb->iconv_out = (iconv_t)-1;
-#endif
-	/* need something to bootstrap with, reassign after page 0 is read */
-	mdb->fmt = &MdbJet3Constants;
-	mdb->f = (MdbFile *) g_malloc0(sizeof(MdbFile));
-	mdb->f->refs = 1;
-	mdb->f->fd = -1;
-	mdb->f->filename = mdb_find_file(filename);
-	if (!mdb->f->filename) { 
+	char *filepath = mdb_find_file(filename);
+	if (!filepath) {
 		fprintf(stderr, "File not found\n");
-		mdb_close(mdb);
 		return NULL; 
 	}
-	if (flags & MDB_WRITABLE) {
-		mdb->f->writable = TRUE;
-		open_flags = O_RDWR;
-	} else {
-		open_flags = O_RDONLY;
-	}
-
 #ifdef _WIN32
-	open_flags |= O_BINARY;
+    char *mode = (flags & MDB_WRITABLE) ? "rb+" : "rb";
+#else
+    char *mode = (flags & MDB_WRITABLE) ? "r+" : "r";
 #endif
 
-	mdb->f->fd = open(mdb->f->filename, open_flags);
-
-	if (mdb->f->fd==-1) {
-		fprintf(stderr,"Couldn't open file %s\n",mdb->f->filename); 
-		mdb_close(mdb);
+    if ((file = fopen(filepath, mode)) == NULL) {
+		fprintf(stderr,"Couldn't open file %s\n",filepath);
+		g_free(filepath);
 		return NULL;
-	}
-	if (!mdb_read_pg(mdb, 0)) {
-		fprintf(stderr,"Couldn't read first page.\n");
-		mdb_close(mdb);
-		return NULL;
-	}
-	if (mdb->pg_buf[0] != 0) {
-		mdb_close(mdb);
-		return NULL; 
-	}
-	mdb->f->jet_version = mdb_get_int32(mdb->pg_buf, 0x14);
-	switch(mdb->f->jet_version) {
-	case MDB_VER_JET3:
-		mdb->fmt = &MdbJet3Constants;
-		break;
-	case MDB_VER_JET4:
-	case MDB_VER_ACCDB_2007:
-	case MDB_VER_ACCDB_2010:
-		mdb->fmt = &MdbJet4Constants;
-		break;
-	default:
-		fprintf(stderr,"Unknown Jet version.\n");
-		mdb_close(mdb);
-		return NULL; 
-	}
-	mdb->f->db_key = mdb_get_int32(mdb->pg_buf, 0x3e);
-	/* I don't know if this value is valid for some versions?
-	 * it doesn't seem to be valid for the databases I have
-	 *
-	 * f->db_key ^= 0xe15e01b9;
-	 */
-	mdb->f->db_key ^= 0x4ebc8afb;
-	/* fprintf(stderr, "Encrypted file, RC4 key seed= %d\n", mdb->f->db_key); */
-	if (mdb->f->db_key) {
-		/* write is not supported for encrypted files yet */
-		mdb->f->writable = FALSE;
-		/* that should be enought, but reopen the file read only just to be
-		 * sure we don't write invalid data */
-		close(mdb->f->fd);
-		open_flags = O_RDONLY;
-#ifdef _WIN32
-		open_flags |= O_BINARY;
-#endif
-		mdb->f->fd = open(mdb->f->filename, open_flags);
-		if (mdb->f->fd==-1) {
-			fprintf(stderr, "Couldn't ropen file %s in read only\n", mdb->f->filename);
-			mdb_close(mdb);
-			return NULL;
-		}
-	}
+    }
 
-	/* get the db password located at 0x42 bytes into the file */
-	for (pos=0;pos<14;pos++) {
-		j = mdb_get_int32(mdb->pg_buf, 0x42+pos);
-		j ^= key[pos];
-		if ( j != 0)
-			mdb->f->db_passwd[pos] = j;
-		else
-			mdb->f->db_passwd[pos] = '\0';
-	}
+    g_free(filepath);
 
-	mdb_iconv_init(mdb);
-
-	return mdb;
+    return mdb_handle_from_stream(file, flags);
 }
 
 /**
@@ -296,13 +330,13 @@ mdb_close(MdbHandle *mdb)
 		if (mdb->f->refs > 1) {
 			mdb->f->refs--;
 		} else {
-			if (mdb->f->fd != -1) close(mdb->f->fd);
-			g_free(mdb->f->filename);
+			if (mdb->f->stream) fclose(mdb->f->stream);
 			g_free(mdb->f);
 		}
 	}
 
 	mdb_iconv_close(mdb);
+    mdb_remove_backends(mdb);
 
 	g_free(mdb);
 }
@@ -322,18 +356,24 @@ MdbHandle *mdb_clone_handle(MdbHandle *mdb)
 	unsigned int i;
 
 	newmdb = (MdbHandle *) g_memdup(mdb, sizeof(MdbHandle));
-	newmdb->stats = NULL;
+
+	memset(&newmdb->catalog, 0, sizeof(MdbHandle) - offsetof(MdbHandle, catalog));
+
 	newmdb->catalog = g_ptr_array_new();
 	for (i=0;i<mdb->num_catalog;i++) {
 		entry = g_ptr_array_index(mdb->catalog,i);
 		data = g_memdup(entry,sizeof(MdbCatalogEntry));
+		data->mdb = newmdb;
+		data->props = NULL;
 		g_ptr_array_add(newmdb->catalog, data);
 	}
-	mdb->backend_name = NULL;
+
+	mdb_iconv_init(newmdb);
+	mdb_set_default_backend(newmdb, mdb->backend_name);
+
 	if (mdb->f) {
 		mdb->f->refs++;
 	}
-	mdb_iconv_init(mdb);
 
 	return newmdb;
 }
@@ -348,7 +388,7 @@ ssize_t mdb_read_pg(MdbHandle *mdb, unsigned long pg)
 	if (pg && mdb->cur_pg == pg) return mdb->fmt->pg_size;
 
 	len = _mdb_read_pg(mdb, mdb->pg_buf, pg);
-	//fprintf(stderr, "read page %d type %02x\n", pg, mdb->pg_buf[0]);
+	//fprintf(stderr, "read page %ld type %02x\n", pg, mdb->pg_buf[0]);
 	mdb->cur_pg = pg;
 	/* kan - reset the cur_pos on a new page read */
 	mdb->cur_pos = 0; /* kan */
@@ -356,35 +396,34 @@ ssize_t mdb_read_pg(MdbHandle *mdb, unsigned long pg)
 }
 ssize_t mdb_read_alt_pg(MdbHandle *mdb, unsigned long pg)
 {
-	ssize_t len;
-
-	len = _mdb_read_pg(mdb, mdb->alt_pg_buf, pg);
-	return len;
+	return _mdb_read_pg(mdb, mdb->alt_pg_buf, pg);
 }
 static ssize_t _mdb_read_pg(MdbHandle *mdb, void *pg_buf, unsigned long pg)
 {
 	ssize_t len;
-	struct stat status;
 	off_t offset = pg * mdb->fmt->pg_size;
 
-        fstat(mdb->f->fd, &status);
-        if (status.st_size < offset) { 
-                fprintf(stderr,"offset %jd is beyond EOF\n",(intmax_t)offset);
-                return 0;
-        }
+    if (fseeko(mdb->f->stream, 0, SEEK_END) == -1) {
+        fprintf(stderr, "Unable to seek to end of file\n");
+        return 0;
+    }
+    if (ftello(mdb->f->stream) < offset) { 
+        fprintf(stderr,"offset %" PRIu64 " is beyond EOF\n",(uint64_t)offset);
+        return 0;
+    }
 	if (mdb->stats && mdb->stats->collect) 
 		mdb->stats->pg_reads++;
 
-	lseek(mdb->f->fd, offset, SEEK_SET);
-	len = read(mdb->f->fd,pg_buf,mdb->fmt->pg_size);
-	if (len==-1) {
+	if (fseeko(mdb->f->stream, offset, SEEK_SET) == -1) {
+        fprintf(stderr, "Unable to seek to page %lu\n", pg);
+        return 0;
+    }
+	len = fread(pg_buf, 1, mdb->fmt->pg_size, mdb->f->stream);
+	if (ferror(mdb->f->stream)) {
 		perror("read");
 		return 0;
 	}
-	else if (len<mdb->fmt->pg_size) {
-		/* fprintf(stderr,"EOF reached %d bytes returned.\n",len, mdb->fmt->pg_size); */
-		return 0;
-	} 
+    memset(pg_buf + len, 0, mdb->fmt->pg_size - len);
 	/*
 	 * unencrypt the page if necessary.
 	 * it might make sense to cache the unencrypted data blocks?
@@ -397,7 +436,7 @@ static ssize_t _mdb_read_pg(MdbHandle *mdb, void *pg_buf, unsigned long pg)
 		RC4(&rc4_key, mdb->fmt->pg_size, pg_buf);
 	}
 
-	return len;
+	return mdb->fmt->pg_size;
 }
 void mdb_swap_pgbuf(MdbHandle *mdb)
 {
@@ -423,7 +462,7 @@ unsigned char mdb_pg_get_byte(MdbHandle *mdb, int offset)
 int mdb_get_int16(void *buf, int offset)
 {
 	guint16 l;
-	memcpy(&l, buf + offset, 2);
+	memcpy(&l, (char*)buf + offset, 2);
 	return (int)GUINT16_FROM_LE(l);
 }
 int mdb_pg_get_int16(MdbHandle *mdb, int offset)
@@ -436,13 +475,13 @@ int mdb_pg_get_int16(MdbHandle *mdb, int offset)
 long mdb_get_int32_msb(void *buf, int offset)
 {
 	gint32 l;
-	memcpy(&l, buf + offset, 4);
+	memcpy(&l, (char*)buf + offset, 4);
 	return (long)GINT32_FROM_BE(l);
 }
 long mdb_get_int32(void *buf, int offset)
 {
 	gint32 l;
-	memcpy(&l, buf + offset, 4);
+	memcpy(&l, (char*)buf + offset, 4);
 	return (long)GINT32_FROM_LE(l);
 }
 long mdb_pg_get_int32(MdbHandle *mdb, int offset)
@@ -455,7 +494,7 @@ long mdb_pg_get_int32(MdbHandle *mdb, int offset)
 float mdb_get_single(void *buf, int offset)
 {
 	union {guint32 g; float f;} f;
-	memcpy(&f, buf + offset, 4);
+	memcpy(&f, (char*)buf + offset, 4);
 	f.g = GUINT32_FROM_LE(f.g);
 	return f.f;
 }
@@ -469,7 +508,7 @@ float mdb_pg_get_single(MdbHandle *mdb, int offset)
 double mdb_get_double(void *buf, int offset)
 {
 	union {guint64 g; double d;} d;
-	memcpy(&d, buf + offset, 8);
+	memcpy(&d, (char*)buf + offset, 8);
 	d.g = GUINT64_FROM_LE(d.g);
 	return d.d;
 }
