@@ -17,13 +17,12 @@
  */
 
 #include "mdbtools.h"
-
-#ifdef DMALLOC
-#include "dmalloc.h"
+#ifdef HAVE_LIBMSWSTR
+#include <mswstr/mswstr.h>
 #endif
 
-MdbIndexPage *mdb_index_read_bottom_pg(MdbHandle *mdb, MdbIndex *idx, MdbIndexChain *chain);
-MdbIndexPage *mdb_chain_add_page(MdbHandle *mdb, MdbIndexChain *chain, guint32 pg);
+static MdbIndexPage *mdb_index_read_bottom_pg(MdbHandle *mdb, MdbIndex *idx, MdbIndexChain *chain);
+static MdbIndexPage *mdb_chain_add_page(MdbHandle *mdb, MdbIndexChain *chain, guint32 pg);
 
 char idx_to_text[] = {
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 0-7     0x00-0x07 */
@@ -60,6 +59,149 @@ char idx_to_text[] = {
 0x81, 0x00, 0x00, 0x00, 'x',  0x00, 0x00, 0x00, /* 0xf8-0xff */
 };
 
+/* This table doesn't really work accurately, as it is missing
+ * a lot of special processing, therefore do not use!
+ * This is just some kind of fallback if MSWSTR cannot be used
+ * for whatever reason and may not work for most indexes, i.e. 
+ * those containing hyphens etc.
+ */
+char idx_to_text_ling[] = {
+0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,  /* 0-7      0x00-0x07 */
+0x01, 0x08, 0x08, 0x08, 0x08, 0x08, 0x01, 0x01,  /* 8-15     0x08-0x0F */
+0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,  /* 16-23    0x10-0x17 */
+0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,  /* 24-31    0x18-0x1F */
+0x07, 0x09, 0x0A, 0x0C, 0x0E, 0x10, 0x12, 0x01,  /* 32-39    0x20-0x27 */
+0x14, 0x16, 0x18,  ',', 0x1A, 0x01, 0x1C, 0x1E,  /* 40-47    0x28-0x2F */
+ '6',  '8',  ':',  '<',  '>',  '@',  'B',  'D',  /* 48-55    0x30-0x37 */
+ 'F',  'H',  ' ',  '"',  '.',  '0',  '2',  '$',  /* 56-63    0x38-0x3F */
+ '&',  'J',  'L',  'M',  'O',  'Q',  'S',  'U',  /* 64-71    0x40-0x47 */
+ 'W',  'Y',  '[',  '\\', '^',  '`',  'b',  'd',  /* 72-79    0x48-0x4F */
+ 'f',  'h',  'i',  'k',  'm',  'o',  'q',  's',  /* 80-87    0x50-0x57 */
+ 'u',  'v',  'x',  '\'', ')',  '*',  '+',  '+',  /* 88-95    0x58-0x5F */
+ '+',  'J',  'L',  'M',  'O',  'Q',  'S',  'U',  /* 96-103   0x60-0x67 */
+ 'W',  'Y',  '[',  '\\', '^',  '`',  'b',  'd',  /* 104-111  0x68-0x6F */
+ 'f',  'h',  'i',  'k',  'm',  'o',  'q',  's',  /* 112-119  0x70-0x77 */
+ 'u',  'v',  'x',  '+',  '+',  '+',  '+', 0x01,  /* 120-127  0x78-0x7F */
+0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,  /* 128-135  0x80-0x87 */
+0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,  /* 136-143  0x88-0x8F */
+0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,  /* 144-151  0x90-0x97 */
+0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,  /* 152-159  0x98-0x9F */
+0x08,  '+',  '4',  '4',  '4',  '4',  '+',  '4',  /* 160-167  0xA0-0xA7 */
+ '+',  '4',  'J',  '3',  '4', 0x01,  '4',  '+',  /* 168-175  0xA8-0xAF */
+ '4',  '3',  ':',  '<',  '+',  '4',  '4',  '4',  /* 176-183  0xB0-0xB7 */
+ '+',  '8',  'd',  '3',  '7',  '7',  '7',  '+',  /* 184-191  0xB8-0xBF */
+ 'J',  'J',  'J',  'J',  'J',  'J',  'J',  'M',  /* 192-199  0xC0-0xC7 */
+ 'Q',  'Q',  'Q',  'Q',  'Y',  'Y',  'Y',  'Y',  /* 200-207  0xC8-0xCF */
+ 'O',  'b',  'd',  'd',  'd',  'd',  'd',  '3',  /* 208-215  0xD0-0xD7 */
+ 'd',  'o',  'o',  'o',  'o',  'v',  'm',  'k',  /* 216-223  0xD8-0xDF */
+ 'J',  'J',  'J',  'J',  'J',  'J',  'J',  'M',  /* 224-231  0xE0-0xE7 */
+ 'Q',  'Q',  'Q',  'Q',  'Y',  'Y',  'Y',  'Y',  /* 232-239  0xE8-0xEF */
+ 'O',  'b',  'd',  'd',  'd',  'd',  'd',  '3',  /* 240-247  0xF0-0xF7 */
+ 'd',  'o',  'o',  'o',  'o',  'v',  'm',  'v',  /* 248-255  0xF8-0xFF */
+};
+
+/* JET Red (v4) Index definition byte layouts
+ *
+ * Based on:
+ *
+ * http://jabakobob.net/mdb/table-page.html
+ * https://github.com/jahlborn/jackcess
+ *
+ * plus inspection of JET (Red) 4 databases.  (JET 3 format has fewer
+ * fields -- some of the ones below omitted, and others narrower.)
+ *
+ * See also JET Blue (Extensible Storage Engine) format information:
+ *
+ * https://github.com/libyal/libesedb/blob/master/documentation/Extensible%20Storage%20Engine%20%28ESE%29%20Database%20File%20%28EDB%29%20format.asciidoc
+ *
+ * which is a later Microsoft embedded database format with the same
+ * early base format.
+ *
+ * ----------------------------------------------------------------------
+ * Index Column Definitions:
+ * - for each "non foreign key" index (ie pidx->index_type!=2), a list
+ *   of columns indexed
+ *
+ * Repeated table->num_real_idxs times:
+ *
+ * Offset	Bytes 	Meaning
+ * 0x0000	4	UNKNOWN; seems to be type marker, usually 1923 or 0
+ *
+ * 0x0004       2       Column 1 ID
+ * 0x0006       1       Column 1 Flags
+ * 0x0007       2       Column 2 ID
+ * 0x0009       1       Column 2 Flags
+ * 0x000A       2       Column 3 ID
+ * 0x000C       1       Column 3 Flags
+ * 0x000D       2       Column 4 ID
+ * 0x000F       1       Column 4 Flags
+ * 0x0010       2       Column 5 ID
+ * 0x0012       1       Column 5 Flags
+ * 0x0013       2       Column 6 ID
+ * 0x0015       1       Column 6 Flags
+ * 0x0016       2       Column 7 ID
+ * 0x0018       1       Column 7 Flags
+ * 0x0019       2       Column 8 ID
+ * 0x001B       1       Column 8 Flags
+ * 0x001C       2       Column 9 ID
+ * 0x001E       1       Column 9 Flags
+ * 0x001F       2       Column 10 ID
+ * 0x0021       1       Column 10 Flags
+ *
+ * 0x0022	1	Usage Map row
+ * 0x0023	3	Usage Map page (24-bit)
+ * 0x0026	4	First index page
+ * 0x002A	4	UNKNOWN
+ * 0x002E	2	Index Flags
+ * 0x0030	4	UNKNOWN; seems to always be 0
+ * 0x0034
+ *
+ * Column ID of 0xFFFF (-1) means "not used" or "end of used columns".
+ * Column Flags:
+ * - 0x01 = Ascending
+ *
+ * Index Flags:
+ * - 0x0001 = Unique index
+ * - 0x0002 = Ignore NULLs
+ * - 0x0008 = Required Index
+ *
+ * ----------------------------------------------------------------------
+ * Index Definitions
+ * - for each index (normal, primary key, foreign key), details on the
+ *   index.
+ *
+ * - this appears to be the union of information required for normal/
+ *   primary key indexes, and the information required for foreign key
+ *   indexes.
+ *
+ * Repeated table->num_idxs times:
+ *
+ * Offset	Bytes 	Meaning
+ * 0x0000	4	UNKNOWN; apparently a type marker, usually 1625 or 0
+ * 0x0004	4	Logical Index Number
+ * 0x0008	4	Index Column Definition Entry
+ * 0x000C	1	FK Index Type
+ * 0x000D	4	FK Index Number
+ * 0x0011	4	FK Index Table Page Number
+ * 0x0015	1	Flags: Update Action
+ * 0x0016	1	Flags: Delete Action
+ * 0x0017	1	Index Type
+ * 0x0018	4	UNKNNOWN; seems to always be 0
+ * 0x001B
+ *
+ * Where Index Type is:
+ * 0x01 = normal
+ * 0x01 = primary key
+ * 0x02 = foreign key index reference
+ */
+
+/* Debugging helper to dump out raw hex values of index definition */
+/*
+static void hexdump(unsigned char *tmpbuf, int size) {
+	mdb_buffer_dump(tmpbuf, 0, size);
+}
+*/
+
 
 GPtrArray *
 mdb_read_indices(MdbTableDef *table)
@@ -86,28 +228,35 @@ mdb_read_indices(MdbTableDef *table)
 		type_offset = 23;
 	}
 
-	//fprintf(stderr, "num_idxs:%d num_real_idxs:%d\n", table->num_idxs, table->num_real_idxs);
-	/* num_real_idxs should be the number of indexes of type 2.
+	/* Read in the definitions of table indexes, into table->indices */
+
+	/* num_real_idxs should be the number of indexes other than type 2.
 	 * It's not always the case. Happens on Northwind Orders table.
 	 */
+	//fprintf(stderr, "num_idxs:%d num_real_idxs:%d\n", table->num_idxs, table->num_real_idxs);
+
 	table->num_real_idxs = 0;
-	tmpbuf = (gchar *) g_malloc(idx2_sz);
+	tmpbuf = g_malloc(idx2_sz);
 	for (i=0;i<table->num_idxs;i++) {
 		read_pg_if_n(mdb, tmpbuf, &cur_pos, idx2_sz);
-		pidx = (MdbIndex *) g_malloc0(sizeof(MdbIndex));
+                //fprintf(stderr, "Index defn: ");
+                //hexdump((unsigned char *)tmpbuf, idx2_sz);
+		pidx = g_malloc0(sizeof(MdbIndex));
 		pidx->table = table;
 		pidx->index_num = mdb_get_int16(tmpbuf, 4);
 		pidx->index_type = tmpbuf[type_offset];
 		g_ptr_array_add(table->indices, pidx);
 		/*
 		{
-			gint32 dumy0 = mdb_get_int32(tmpbuf, 0);
-			gint8 dumy1 = tmpbuf[8];
-			gint32 dumy2 = mdb_get_int32(tmpbuf, 9);
-			gint32 dumy3 = mdb_get_int32(tmpbuf, 13);
-			gint16 dumy4 = mdb_get_int16(tmpbuf, 17);
-			fprintf(stderr, "idx #%d: num2:%d type:%d\n", i, pidx->index_num, pidx->index_type);
-			fprintf(stderr, "idx #%d: %d %d %d %d %d\n", i, dumy0, dumy1, dumy2, dumy3, dumy4);
+			gint32 idx_marker = mdb_get_int32(tmpbuf, 0);
+		        gint32 index_col_def_num = mdb_get_int16(tmpbuf, 8);
+			gint8 rel_idx_type = tmpbuf[0x0c];
+			gint32 rel_idx_number = mdb_get_int32(tmpbuf, 0x0d);
+			gint32 rel_idx_page = mdb_get_int32(tmpbuf, 0x11);
+			gint8 update_action_flags = tmpbuf[0x15];
+			gint8 delete_action_flags = tmpbuf[0x16];
+			fprintf(stderr, "idx #%d: num2:%d num3:%d type:%d\n", i, pidx->index_num, index_col_def_num, pidx->index_type);
+			fprintf(stderr, "idx #%d: %d %d %d %d %d/%d\n", i, idx_marker, rel_idx_type, rel_idx_number, rel_idx_page, update_action_flags, delete_action_flags);
 		}*/
 		if (pidx->index_type!=2)
 			table->num_real_idxs++;
@@ -115,6 +264,7 @@ mdb_read_indices(MdbTableDef *table)
 	//fprintf(stderr, "num_idxs:%d num_real_idxs:%d\n", table->num_idxs, table->num_real_idxs);
 	g_free(tmpbuf);
 
+	/* Pick up the names of each index */
 	for (i=0;i<table->num_idxs;i++) {
 		pidx = g_ptr_array_index (table->indices, i);
 		if (IS_JET3(mdb)) {
@@ -124,20 +274,35 @@ mdb_read_indices(MdbTableDef *table)
 		}
 		tmpbuf = g_malloc(name_sz);
 		read_pg_if_n(mdb, tmpbuf, &cur_pos, name_sz);
-		mdb_unicode2ascii(mdb, tmpbuf, name_sz, pidx->name, MDB_MAX_OBJ_NAME); 
+		mdb_unicode2ascii(mdb, tmpbuf, name_sz, pidx->name, sizeof(pidx->name));
 		g_free(tmpbuf);
 		//fprintf(stderr, "index %d type %d name %s\n", pidx->index_num, pidx->index_type, pidx->name);
 	}
 
+	/* Pick up the column definitions for normal/primary key indexes */
+	/* NOTE: Match should possibly be by index_col_def_num, rather
+         * than index_num; but in files encountered both seem to be the
+         * same (so left with index_num until a counter example is found).
+         */
 	mdb_read_alt_pg(mdb, entry->table_pg);
 	mdb_read_pg(mdb, index_start_pg);
 	cur_pos = table->index_start;
 	for (i=0;i<table->num_real_idxs;i++) {
+		/*	Debugging print out, commented out
+		{
+			gchar *tmpbuf = (gchar *) g_malloc(0x34);
+			int now_pos = cur_pos;
+			read_pg_if_n(mdb, tmpbuf, &now_pos, 0x34);
+			fprintf(stderr, "Index defn: ");
+			hexdump((unsigned char *)tmpbuf, 0x34);
+			g_free(tmpbuf);
+		}*/
+
 		if (!IS_JET3(mdb)) cur_pos += 4;
 		/* look for index number i */
 		for (j=0; j<table->num_idxs; ++j) {
 			pidx = g_ptr_array_index (table->indices, j);
-			if (pidx->index_type!=2 && pidx->index_num==i)
+			if (pidx->index_type!=2 && (unsigned int)pidx->index_num==i)
 				break;
 		}
 		if (j==table->num_idxs) {
@@ -152,12 +317,13 @@ mdb_read_indices(MdbTableDef *table)
 		/*
 		fprintf(stderr, "ridx block1 i:%d data1:0x%08x data2:0x%08x\n",
 			i,
-			mdb_get_int32(mdb->pg_buf,
+			(unsigned int)mdb_get_int32(mdb->pg_buf,
 				fmt->tab_cols_start_offset + pidx->index_num * fmt->tab_ridx_entry_size),
-			mdb_get_int32(mdb->pg_buf,
+			(unsigned int)mdb_get_int32(mdb->pg_buf,
 				fmt->tab_cols_start_offset + pidx->index_num * fmt->tab_ridx_entry_size +4));
 		fprintf(stderr, "pidx->num_rows:%d\n", pidx->num_rows);*/
 
+		/* Read columns in each index */
 		key_num=0;
 		for (j=0;j<MDB_MAX_IDX_COLS;j++) {
 			col_num=read_pg_if_16(mdb,&cur_pos);
@@ -190,27 +356,65 @@ mdb_read_indices(MdbTableDef *table)
 		}
 		pidx->num_keys = key_num;
 
-		cur_pos += 4;
-		//fprintf(stderr, "pidx->unknown_pre_first_pg:0x%08x\n", read_pg_if_32(mdb, &cur_pos));
+		if (0)  // DEBUGGING ONLY
+		{
+			gint32 usage_map = read_pg_if_32(mdb, &cur_pos);
+			fprintf(stderr, "pidx->unknown_pre_first_pg:0x%08x\n", usage_map);
+		} else {
+			cur_pos += 4;    // Skip Usage map information
+		}
 		pidx->first_pg = read_pg_if_32(mdb, &cur_pos);
+
+		if (!IS_JET3(mdb)) cur_pos += 4;
+
 		pidx->flags = read_pg_if_8(mdb, &cur_pos);
 		//fprintf(stderr, "pidx->first_pg:%d pidx->flags:0x%02x\n",	pidx->first_pg, pidx->flags);
-		if (!IS_JET3(mdb)) cur_pos += 9;
+		if (!IS_JET3(mdb)) cur_pos += 5;
 	}
 	return NULL;
 }
 void
-mdb_index_hash_text(char *text, char *hash)
+mdb_index_hash_text(MdbHandle *mdb, char *text, char *hash)
 {
-	unsigned int k;
+	unsigned int k, len=strlen(text);
+	char *transtbl=NULL;
 
-	for (k=0;k<strlen(text);k++) {
-		int c = ((unsigned char *)(text))[k];
-		hash[k] = idx_to_text[c];
-		if (!(hash[k])) fprintf(stderr, 
-				"No translation available for %02x %d\n", c, c);
+	if (!IS_JET3(mdb))
+	{
+#ifdef __MSWSTR_H__
+		char *out_ptr = alloca((len+1)*2);
+		unsigned int i;
+		// mdb_ascii2unicode doesn't work, we don't want unicode compression!
+		for (i=0; i<len+1; i++) {
+			out_ptr[i*2] = text[i];
+			out_ptr[i*2+1] = 0;
+		}
+        if (!(k=DBLCMapStringW(MAKELCID(mdb->f->lang_id, 0),
+                LCMAP_LINGUISTIC_CASING | LCMAP_SORTKEY | NORM_IGNORECASE | NORM_IGNOREKANATYPE | NORM_IGNOREWIDTH,
+                (WCHAR*)out_ptr, len, (LPBYTE)hash, len*2)))
+		{
+			len++;
+#endif
+		transtbl = idx_to_text_ling;
+#ifdef __MSWSTR_H__
+		}
+#endif
 	}
-	hash[strlen(text)]='\0';
+	else
+	{
+		transtbl = idx_to_text;
+	}
+	if (transtbl)
+	{
+		for (k=0;k<len;k++) {
+			unsigned char c = ((unsigned char *)(text))[k];
+			hash[k] = transtbl[c];
+			if (!(hash[k])) fprintf(stderr, 
+					"No translation available for %02x %d\n", c, c);
+		}
+		hash[len]='\0';
+	}
+	//printf ("mdb_index_hash_text %s -> %s (%d -> %d)\n", text, hash, len, k);
 }
 /*
  * reverse the order of the column for hashing
@@ -232,7 +436,7 @@ mdb_index_cache_sarg(MdbColumn *col, MdbSarg *sarg, MdbSarg *idx_sarg)
 
 	switch (col->col_type) {
 		case MDB_TEXT:
-		mdb_index_hash_text(sarg->value.s, idx_sarg->value.s);
+		mdb_index_hash_text(col->table->mdbidx, sarg->value.s, idx_sarg->value.s);
 		break;
 
 		case MDB_LONGINT:
@@ -293,8 +497,7 @@ mdb_index_test_sargs(MdbHandle *mdb, MdbIndex *idx, char *buf, int len)
 	int c_len;
 
 	//fprintf(stderr,"mdb_index_test_sargs called on ");
-	//for (i=0;i<len;i++)
-		//fprintf(stderr,"%02x ",buf[i]); //mdb->pg_buf[offset+i]);
+	//mdb_buffer_dump(buf, 0, len);
 	//fprintf(stderr,"\n");
 	for (i=0;i<idx->num_keys;i++) {
 		//c_offset++; /* the per column null indicator/flags */
@@ -333,7 +536,12 @@ mdb_index_test_sargs(MdbHandle *mdb, MdbIndex *idx, char *buf, int len)
 			field.value = buf;
 		       	field.siz = c_len;
 		       	field.is_null = FALSE;
-			if (!mdb_test_sarg(mdb, col, &node, &field)) {
+			/* In Jet 4 Index text hashes don't need to be converted from Unicode */
+			if (!IS_JET3(mdb) && col->col_type == MDB_TEXT)
+			{
+				if (!mdb_test_string(&node, buf)) return 0;
+			}
+			else if (!mdb_test_sarg(mdb, col, &node, &field)) {
 				/* sarg didn't match, no sense going on */
 				return 0;
 			}
@@ -348,7 +556,7 @@ int
 mdb_index_pack_bitmap(MdbHandle *mdb, MdbIndexPage *ipg)
 {
 	int mask_bit = 0;
-	int mask_pos = 0x16;
+	int mask_pos = IS_JET3(mdb)?0x16:0x1b;
 	int mask_byte = 0;
 	int elem = 0;
 	int len, start, i;
@@ -387,9 +595,10 @@ int
 mdb_index_unpack_bitmap(MdbHandle *mdb, MdbIndexPage *ipg)
 {
 	int mask_bit = 0;
-	int mask_pos = 0x16;
+	int mask_pos = IS_JET3(mdb)?0x16:0x1b;
 	int mask_byte;
-	int start = 0xf8;
+	int jet_start = IS_JET3(mdb)?0xf8:0x1e0;
+	int start = jet_start;
 	int elem = 0;
 	int len = 0;
 
@@ -406,13 +615,13 @@ mdb_index_unpack_bitmap(MdbHandle *mdb, MdbIndexPage *ipg)
 			}
 			mask_byte = mdb->pg_buf[mask_pos];
 			len++;
-		} while (mask_pos <= 0xf8 && !((1 << mask_bit) & mask_byte));
+		} while (mask_pos <= jet_start && !((1 << mask_bit) & mask_byte));
 		//fprintf(stdout, "%d %d %d %d\n", mask_pos, mask_bit, mask_byte, len);
 
 		start += len;
-		if (mask_pos < 0xf8) ipg->idx_starts[elem++]=start;
+		if (mask_pos < jet_start) ipg->idx_starts[elem++]=start;
 
-	} while (mask_pos < 0xf8);
+	} while (mask_pos < jet_start);
 
 	/* if we zero the next element, so we don't pick up the last pages starts*/
 	ipg->idx_starts[elem]=0;
@@ -442,17 +651,17 @@ mdb_index_find_next_on_page(MdbHandle *mdb, MdbIndexPage *ipg)
 
 	return ipg->len;
 }
-void mdb_index_page_reset(MdbIndexPage *ipg)
+void mdb_index_page_reset(MdbHandle *mdb, MdbIndexPage *ipg)
 {
-	ipg->offset = 0xf8; /* start byte of the index entries */
+	ipg->offset = IS_JET3(mdb)?0xf8:0x1e0; /* start byte of the index entries */
 	ipg->start_pos=0;
 	ipg->len = 0; 
 	ipg->idx_starts[0]=0;
 }
-void mdb_index_page_init(MdbIndexPage *ipg)
+void mdb_index_page_init(MdbHandle *mdb, MdbIndexPage *ipg)
 {
 	memset(ipg, 0, sizeof(MdbIndexPage));
-	mdb_index_page_reset(ipg);
+	mdb_index_page_reset(mdb, ipg);
 }
 /*
  * find the next leaf page if any given a chain. Assumes any exhausted leaf 
@@ -498,7 +707,8 @@ mdb_find_next_leaf(MdbHandle *mdb, MdbIndex *idx, MdbIndexChain *chain)
 		 * add to the chain and call this function
 		 * recursively.
 		 */
-		newipg = mdb_chain_add_page(mdb, chain, pg);
+		if (!mdb_chain_add_page(mdb, chain, pg))
+			break;
 		newipg = mdb_find_next_leaf(mdb, idx, chain);
 		//printf("returning pg %lu\n",newipg->pg);
 		return newipg;
@@ -507,7 +717,7 @@ mdb_find_next_leaf(MdbHandle *mdb, MdbIndex *idx, MdbIndexChain *chain)
 	return NULL;
 
 }
-MdbIndexPage *
+static MdbIndexPage *
 mdb_chain_add_page(MdbHandle *mdb, MdbIndexChain *chain, guint32 pg)
 {
 	MdbIndexPage *ipg;
@@ -515,10 +725,10 @@ mdb_chain_add_page(MdbHandle *mdb, MdbIndexChain *chain, guint32 pg)
 	chain->cur_depth++;
 	if (chain->cur_depth > MDB_MAX_INDEX_DEPTH) {
 		fprintf(stderr,"Error! maximum index depth of %d exceeded.  This is probably due to a programming bug, If you are confident that your indexes really are this deep, adjust MDB_MAX_INDEX_DEPTH in mdbtools.h and recompile.\n", MDB_MAX_INDEX_DEPTH);
-		exit(1);
+		return NULL;
 	}
 	ipg = &(chain->pages[chain->cur_depth - 1]);
-	mdb_index_page_init(ipg);
+	mdb_index_page_init(mdb, ipg);
 	ipg->pg = pg;
 
 	return ipg;
@@ -527,7 +737,7 @@ mdb_chain_add_page(MdbHandle *mdb, MdbIndexChain *chain, guint32 pg)
  * returns the bottom page of the IndexChain, if IndexChain is empty it 
  * initializes it by reading idx->first_pg (the root page)
  */
-MdbIndexPage *
+static MdbIndexPage *
 mdb_index_read_bottom_pg(MdbHandle *mdb, MdbIndex *idx, MdbIndexChain *chain)
 {
 	MdbIndexPage *ipg;
@@ -537,7 +747,7 @@ mdb_index_read_bottom_pg(MdbHandle *mdb, MdbIndex *idx, MdbIndexChain *chain)
 	 */
 	if (!chain->cur_depth) {
 		ipg = &(chain->pages[0]);
-		mdb_index_page_init(ipg);
+		mdb_index_page_init(mdb, ipg);
 		chain->cur_depth = 1;
 		ipg->pg = idx->first_pg;
 		if (!(ipg = mdb_find_next_leaf(mdb, idx, chain)))
@@ -600,6 +810,7 @@ mdb_index_find_next(MdbHandle *mdb, MdbIndex *idx, MdbIndexChain *chain, guint32
 	int passed = 0;
 	int idx_sz;
 	int idx_start = 0;
+	unsigned short compress_bytes;
 	MdbColumn *col;
 	guint32 pg_row;
 
@@ -615,7 +826,7 @@ mdb_index_find_next(MdbHandle *mdb, MdbIndex *idx, MdbIndexChain *chain, guint32
 		 */
 		if (!mdb_index_find_next_on_page(mdb, ipg)) {
 			if (!chain->clean_up_mode) {
-				if (!(ipg = mdb_index_unwind(mdb, idx, chain)))
+				if (ipg->rc==1 || !(ipg = mdb_index_unwind(mdb, idx, chain)))
 					chain->clean_up_mode = 1;
 			}
 			if (chain->clean_up_mode) {
@@ -630,7 +841,7 @@ mdb_index_find_next(MdbHandle *mdb, MdbIndex *idx, MdbIndexChain *chain, guint32
 				/* reuse the chain for cleanup mode */
 				chain->cur_depth = 1;
 				ipg = &chain->pages[0];
-				mdb_index_page_init(ipg);
+				mdb_index_page_init(mdb, ipg);
 				ipg->pg = chain->last_leaf_found;
 				//printf("next on page %d\n",
 				if (!mdb_index_find_next_on_page(mdb, ipg))
@@ -644,10 +855,12 @@ mdb_index_find_next(MdbHandle *mdb, MdbIndex *idx, MdbIndexChain *chain, guint32
 		col=g_ptr_array_index(idx->table->columns,idx->key_col_num[0]-1);
 		idx_sz = mdb_col_fixed_size(col);
 		/* handle compressed indexes, single key indexes only? */
-		if (idx->num_keys==1 && idx_sz>0 && ipg->len - 4 < idx_sz) {
+		if (idx_sz<0) idx_sz = ipg->len - (ipg->start_pos==1?5:4); // Length from Index - the 4 trailing bytes (data page/row), Skip flags on first page
+		compress_bytes = mdb_get_int16(mdb->pg_buf, IS_JET3(mdb)?0x14:0x18);
+		if (idx->num_keys==1 && idx_sz>0 && compress_bytes > 1 && ipg->start_pos>1 /*ipg->len - 4 < idx_sz*/) {
 			//printf("short index found\n");
 			//mdb_buffer_dump(ipg->cache_value, 0, idx_sz);
-			memcpy(&ipg->cache_value[idx_sz - (ipg->len - 4)], &mdb->pg_buf[ipg->offset], ipg->len);
+			memcpy(&ipg->cache_value[compress_bytes-1], &mdb->pg_buf[ipg->offset], ipg->len);
 			//mdb_buffer_dump(ipg->cache_value, 0, idx_sz);
 		} else {
 			idx_start = ipg->offset + (ipg->len - 4 - idx_sz);
@@ -656,6 +869,7 @@ mdb_index_find_next(MdbHandle *mdb, MdbIndex *idx, MdbIndexChain *chain, guint32
 
 		//idx_start = ipg->offset + (ipg->len - 4 - idx_sz);
 		passed = mdb_index_test_sargs(mdb, idx, (char *)(ipg->cache_value), idx_sz);
+		if (passed) ipg->rc=1; else if (ipg->rc) return 0;
 
 		ipg->offset += ipg->len;
 	} while (!passed);
@@ -915,7 +1129,7 @@ mdb_index_scan_free(MdbTableDef *table)
 
 void mdb_free_indices(GPtrArray *indices)
 {
-	unsigned int i;
+	guint i;
 
 	if (!indices) return;
 	for (i=0; i<indices->len; i++)
