@@ -22,16 +22,12 @@
 #include <stdlib.h>
 #include "mdbtools.h"
 
-#ifdef DMALLOC
-#include "dmalloc.h"
-#endif
-
 #define DEBUG 1
 
-static unsigned long opts;
-static int optset;
+static __thread unsigned long opts;
+static __thread int optset;
 
-static void load_options();
+static void load_options(void);
 
 void
 mdb_debug(int klass, char *fmt, ...)
@@ -54,11 +50,18 @@ load_options()
 {
 	char *opt;
 	char *s;
+    char *ctx;
 
     if (!optset && (s=getenv("MDBOPTS"))) {
-		opt = strtok(s, ":");
+		opt = strtok_r(s, ":", &ctx);
 		while (opt) {
-        	if (!strcmp(opt, "use_index")) opts |= MDB_USE_INDEX;
+			if (!strcmp(opt, "use_index")) {
+#ifdef HAVE_LIBMSWSTR
+				opts |= MDB_USE_INDEX;
+#else
+				fprintf(stderr, "The 'use_index' argument was supplied to MDBOPTS environment variable. However, this feature requires the libmswstr library, which was not found when libmdb was compiled. As a result, the 'use_index' argument will be ignored.\n\nTo enable indexes, you will need to download libmswstr from https://github.com/leecher1337/libmswstr and then recompile libmdb. Note that the 'use_index' feature is largely untested, and may have unexpected results.\n\nTo suppress this warning, run the program again after removing the 'use_index' argument from the MDBOPTS environment variable.\n");
+#endif
+			}
         	if (!strcmp(opt, "no_memo")) opts |= MDB_NO_MEMO;
         	if (!strcmp(opt, "debug_like")) opts |= MDB_DEBUG_LIKE;
         	if (!strcmp(opt, "debug_write")) opts |= MDB_DEBUG_WRITE;
@@ -74,7 +77,7 @@ load_options()
 				opts |= MDB_DEBUG_ROW;
 				opts |= MDB_DEBUG_PROPS;
 			}
-			opt = strtok(NULL,":");
+			opt = strtok_r(NULL,":", &ctx);
 		}
     }
 	optset = 1;
