@@ -40,7 +40,7 @@ mdb_read_props_list(MdbHandle *mdb, gchar *kkd, int len)
 			mdb_buffer_dump(kkd, pos - 2, record_len + 2);
 		}
 		name = g_malloc(3*record_len + 1); /* worst case scenario is 3 bytes out per byte in */
-		mdb_unicode2ascii(mdb, &kkd[pos], record_len, name, 3*record_len);
+		mdb_unicode2ascii(mdb, &kkd[pos], record_len, name, 3*record_len + 1);
 
 		pos += record_len;
 		g_ptr_array_add(names, name);
@@ -76,7 +76,7 @@ free_names(GPtrArray *names) {
 	g_ptr_array_free(names, TRUE);
 }
 MdbProperties *
-mdb_alloc_props()
+mdb_alloc_props(void)
 {
 	MdbProperties *props;
 
@@ -89,7 +89,8 @@ mdb_read_props(MdbHandle *mdb, GPtrArray *names, gchar *kkd, int len)
 {
 	guint32 record_len, name_len;
 	int pos = 0;
-	int elem, dtype, dsize;
+	guint elem;
+	int dtype, dsize;
 	gchar *name, *value;
 	MdbProperties *props;
 	int i=0;
@@ -106,7 +107,7 @@ mdb_read_props(MdbHandle *mdb, GPtrArray *names, gchar *kkd, int len)
 	props = mdb_alloc_props();
 	if (name_len) {
 		props->name = g_malloc(3*name_len + 1);
-		mdb_unicode2ascii(mdb, kkd+pos, name_len, props->name, 3*name_len);
+		mdb_unicode2ascii(mdb, kkd+pos, name_len, props->name, 3*name_len + 1);
 		mdb_debug(MDB_DEBUG_PROPS,"prop block named: %s", props->name);
 	}
 	pos += name_len;
@@ -117,10 +118,12 @@ mdb_read_props(MdbHandle *mdb, GPtrArray *names, gchar *kkd, int len)
 		record_len = mdb_get_int16(kkd, pos);
 		dtype = kkd[pos + 3];
 		elem = mdb_get_int16(kkd, pos + 4);
+		if (elem < 0 || elem >= names->len)
+			break;
 		dsize = mdb_get_int16(kkd, pos + 6);
-		value = g_malloc(dsize + 1);
-		strncpy(value, &kkd[pos + 8], dsize);
-		value[dsize] = '\0';
+		if (dsize < 0 || pos + 8 + dsize > len)
+			break;
+		value = g_strdup_printf("%*s", dsize, &kkd[pos+8]);
 		name = g_ptr_array_index(names,elem);
 		if (mdb_get_option(MDB_DEBUG_PROPS)) {
 			fprintf(stderr, "%02d ",i++);
@@ -131,6 +134,9 @@ mdb_read_props(MdbHandle *mdb, GPtrArray *names, gchar *kkd, int len)
 		if (dtype == MDB_BOOL) {
 			g_hash_table_insert(props->hash, g_strdup(name),
 				g_strdup(kkd[pos + 8] ? "yes" : "no"));
+        } else if (dtype == MDB_BINARY && dsize == 16 && strcmp(name, "GUID") == 0) {
+            gchar *guid = mdb_uuid_to_string(kkd, pos+8);
+			g_hash_table_insert(props->hash, g_strdup(name), guid);
 		} else {
 			g_hash_table_insert(props->hash, g_strdup(name),
 			  mdb_col_to_string(mdb, kkd, pos + 8, dtype, dsize));
@@ -159,16 +165,16 @@ mdb_dump_props(MdbProperties *props, FILE *outfile, int show_name) {
 /*
  * That function takes a raw KKD/MR2 binary buffer,
  * typically read from LvProp in table MSysbjects
- * and returns a GArray of MdbProps*
+ * and returns a GPtrArray of MdbProps*
  */
-GArray*
+GPtrArray*
 mdb_kkd_to_props(MdbHandle *mdb, void *buffer, size_t len) {
 	guint32 record_len;
 	guint16 record_type;
 	size_t pos;
 	GPtrArray *names = NULL;
 	MdbProperties *props;
-	GArray *result;
+	GPtrArray *result;
 
 #if MDB_DEBUG
 	mdb_buffer_dump(buffer, 0, len);
@@ -181,7 +187,7 @@ mdb_kkd_to_props(MdbHandle *mdb, void *buffer, size_t len) {
 		return NULL;
 	}
 
-	result = g_array_new(0, 0, sizeof(MdbProperties*));
+	result = g_ptr_array_new();
 
 	pos = 4;
 	while (pos < len) {
@@ -192,16 +198,17 @@ mdb_kkd_to_props(MdbHandle *mdb, void *buffer, size_t len) {
 		switch (record_type) {
 			case 0x80:
 				if (names) free_names(names);
-				names = mdb_read_props_list(mdb, buffer+pos+6, record_len - 6);
+				names = mdb_read_props_list(mdb, (char*)buffer+pos+6, record_len - 6);
 				break;
 			case 0x00:
 			case 0x01:
+			case 0x02:
 				if (!names) {
 					fprintf(stderr,"sequence error!\n");
 					break;
 				}
-				props = mdb_read_props(mdb, names, buffer+pos+6, record_len - 6);
-				g_array_append_val(result, props);
+				props = mdb_read_props(mdb, names, (char*)buffer+pos+6, record_len - 6);
+				g_ptr_array_add(result, props);
 				//mdb_dump_props(props, stderr, 1);
 				break;
 			default:
