@@ -18,11 +18,6 @@
 
 #include "mdbtools.h"
 
-#ifdef DMALLOC
-#include "dmalloc.h"
-#endif
-
-
 static gint mdb_col_comparer(MdbColumn **a, MdbColumn **b)
 {
 	if ((*a)->col_num > (*b)->col_num)
@@ -33,22 +28,11 @@ static gint mdb_col_comparer(MdbColumn **a, MdbColumn **b)
 		return 0;
 }
 
-unsigned char mdb_col_needs_size(int col_type)
-{
-	if (col_type == MDB_TEXT) {
-		return TRUE;
-	} else {
-		return FALSE;
-	}
-}
-
 MdbTableDef *mdb_alloc_tabledef(MdbCatalogEntry *entry)
 {
-	MdbTableDef *table;
-
-	table = (MdbTableDef *) g_malloc0(sizeof(MdbTableDef));
+	MdbTableDef *table = g_malloc0(sizeof(MdbTableDef));
 	table->entry=entry;
-	strcpy(table->name, entry->object_name);
+	snprintf(table->name, sizeof(table->name), "%s", entry->object_name);
 
 	return table;	
 }
@@ -56,7 +40,7 @@ void mdb_free_tabledef(MdbTableDef *table)
 {
 	if (!table) return;
 	if (table->is_temp_table) {
-		unsigned int i;
+		guint i;
 		/* Temp table pages are being stored in memory */
 		for (i=0; i<table->temp_table_pages->len; i++)
 			g_free(g_ptr_array_index(table->temp_table_pages,i));
@@ -79,13 +63,21 @@ MdbTableDef *mdb_read_table(MdbCatalogEntry *entry)
 	void *buf, *pg_buf = mdb->pg_buf;
 	guint i;
 
-	mdb_read_pg(mdb, entry->table_pg);
-	if (mdb_get_byte(pg_buf, 0) != 0x02)  /* not a valid table def page */
+	if (!mdb_read_pg(mdb, entry->table_pg)) {
+        fprintf(stderr, "mdb_read_table: Unable to read page %lu\n", entry->table_pg);
+        return NULL;
+    }
+	if (mdb_get_byte(pg_buf, 0) != 0x02) {
+        fprintf(stderr, "mdb_read_table: Page %lu [size=%d] is not a valid table definition page (First byte = 0x%02X, expected 0x02)\n",
+                entry->table_pg, (int)fmt->pg_size, mdb_get_byte(pg_buf, 0));
 		return NULL;
+    }
 	table = mdb_alloc_tabledef(entry);
 
 	mdb_get_int16(pg_buf, 8); /* len */
 
+	/* Note that num_rows may be zero if the database was improperly closed.
+	 * See https://github.com/mdbtools/mdbtools/issues/120 for discussion. */
 	table->num_rows = mdb_get_int32(pg_buf, fmt->tab_num_rows_offset);
 	table->num_var_cols = mdb_get_int16(pg_buf, fmt->tab_num_cols_offset-2);
 	table->num_cols = mdb_get_int16(pg_buf, fmt->tab_num_cols_offset);
@@ -94,8 +86,12 @@ MdbTableDef *mdb_read_table(MdbCatalogEntry *entry)
 
 	/* grab a copy of the usage map */
 	pg_row = mdb_get_int32(pg_buf, fmt->tab_usage_map_offset);
-	mdb_find_pg_row(mdb, pg_row, &buf, &row_start, &(table->map_sz));
-	table->usage_map = g_memdup(buf + row_start, table->map_sz);
+	if (mdb_find_pg_row(mdb, pg_row, &buf, &row_start, &(table->map_sz))) {
+        fprintf(stderr, "mdb_read_table: Unable to find page row %d\n", pg_row);
+		mdb_free_tabledef(table);
+		return NULL;
+	}
+	table->usage_map = g_memdup((char*)buf + row_start, table->map_sz);
 	if (mdb_get_option(MDB_DEBUG_USAGE)) 
 		mdb_buffer_dump(buf, row_start, table->map_sz);
 	mdb_debug(MDB_DEBUG_USAGE,"usage map found on page %ld row %d start %d len %d",
@@ -103,8 +99,12 @@ MdbTableDef *mdb_read_table(MdbCatalogEntry *entry)
 
 	/* grab a copy of the free space page map */
 	pg_row = mdb_get_int32(pg_buf, fmt->tab_free_map_offset);
-	mdb_find_pg_row(mdb, pg_row, &buf, &row_start, &(table->freemap_sz));
-	table->free_usage_map = g_memdup(buf + row_start, table->freemap_sz);
+	if (mdb_find_pg_row(mdb, pg_row, &buf, &row_start, &(table->freemap_sz))) {
+        fprintf(stderr, "mdb_read_table: Unable to find page row %d\n", pg_row);
+		mdb_free_tabledef(table);
+		return NULL;
+	}
+	table->free_usage_map = g_memdup((char*)buf + row_start, table->freemap_sz);
 	mdb_debug(MDB_DEBUG_USAGE,"free map found on page %ld row %d start %d len %d\n",
 		pg_row >> 8, pg_row & 0xff, row_start, table->freemap_sz);
 
@@ -112,7 +112,7 @@ MdbTableDef *mdb_read_table(MdbCatalogEntry *entry)
 
 	if (entry->props)
 		for (i=0; i<entry->props->len; ++i) {
-			MdbProperties *props = g_array_index(entry->props, MdbProperties*, i);
+			MdbProperties *props = g_ptr_array_index(entry->props, i);
 			if (!props->name)
 				table->props = props;
 		}
@@ -128,7 +128,7 @@ MdbTableDef *mdb_read_table_by_name(MdbHandle *mdb, gchar *table_name, int obj_t
 
 	for (i=0; i<mdb->num_catalog; i++) {
 		entry = g_ptr_array_index(mdb->catalog, i);
-		if (!strcasecmp(entry->object_name, table_name))
+		if (!g_ascii_strcasecmp(entry->object_name, table_name))
 			return mdb_read_table(entry);
 	}
 
@@ -168,28 +168,30 @@ read_pg_if_8(MdbHandle *mdb, int *cur_pos)
 void * 
 read_pg_if_n(MdbHandle *mdb, void *buf, int *cur_pos, size_t len)
 {
+	char* _buf = buf;
+
 	/* Advance to page which contains the first byte */
 	while (*cur_pos >= mdb->fmt->pg_size) {
 		mdb_read_pg(mdb, mdb_get_int32(mdb->pg_buf,4));
 		*cur_pos -= (mdb->fmt->pg_size - 8);
 	}
 	/* Copy pages into buffer */
-	while (*cur_pos + len >= mdb->fmt->pg_size) {
+	while (*cur_pos + len >= (size_t)mdb->fmt->pg_size) {
 		int piece_len = mdb->fmt->pg_size - *cur_pos;
-		if (buf) {
-			memcpy(buf, mdb->pg_buf + *cur_pos, piece_len);
-			buf += piece_len;
+		if (_buf) {
+			memcpy(_buf, mdb->pg_buf + *cur_pos, piece_len);
+			_buf += piece_len;
 		}
 		len -= piece_len;
 		mdb_read_pg(mdb, mdb_get_int32(mdb->pg_buf,4));
 		*cur_pos = 8;
 	}
 	/* Copy into buffer from final page */
-	if (len && buf) {
-		memcpy(buf, mdb->pg_buf + *cur_pos, len);
+	if (len && _buf) {
+		memcpy(_buf, mdb->pg_buf + *cur_pos, len);
 	}
 	*cur_pos += len;
-	return buf;
+	return _buf;
 }
 
 
@@ -199,7 +201,7 @@ void mdb_append_column(GPtrArray *columns, MdbColumn *in_col)
 }
 void mdb_free_columns(GPtrArray *columns)
 {
-	unsigned int i, j;
+	guint i, j;
 	MdbColumn *col;
 
 	if (!columns) return;
@@ -221,14 +223,15 @@ GPtrArray *mdb_read_columns(MdbTableDef *table)
 	MdbFormatConstants *fmt = mdb->fmt;
 	MdbColumn *pcol;
 	unsigned char *col;
-	unsigned int i, j;
+	unsigned int i;
+	guint j;
 	int cur_pos;
 	size_t name_sz;
-	GArray *allprops;
+	GPtrArray *allprops;
 	
 	table->columns = g_ptr_array_new();
 
-	col = (unsigned char *) g_malloc(fmt->tab_col_entry_size);
+	col = g_malloc(fmt->tab_col_entry_size);
 
 	cur_pos = fmt->tab_cols_start_offset + 
 		(table->num_real_idxs * fmt->tab_ridx_entry_size);
@@ -244,7 +247,7 @@ GPtrArray *mdb_read_columns(MdbTableDef *table)
 	mdb_buffer_dump(mdb->pg_buf, cur_pos, fmt->tab_col_entry_size); */
 #endif
 		read_pg_if_n(mdb, col, &cur_pos, fmt->tab_col_entry_size);
-		pcol = (MdbColumn *) g_malloc0(sizeof(MdbColumn));
+		pcol = g_malloc0(sizeof(MdbColumn));
 
 		pcol->table = table;
 
@@ -264,8 +267,8 @@ GPtrArray *mdb_read_columns(MdbTableDef *table)
 
 		/* FIXME: can this be right in Jet3 and Jet4? */
 		if (pcol->col_type == MDB_NUMERIC) {
-			pcol->col_prec = col[11];
-			pcol->col_scale = col[12];
+			pcol->col_scale = col[11];
+			pcol->col_prec = col[12];
 		}
 
 		// col_flags_offset == 13 or 15
@@ -301,12 +304,10 @@ GPtrArray *mdb_read_columns(MdbTableDef *table)
 			name_sz = read_pg_if_8(mdb, &cur_pos);
 		else
 			name_sz = read_pg_if_16(mdb, &cur_pos);
-		tmp_buf = (char *) g_malloc(name_sz);
+		tmp_buf = g_malloc(name_sz);
 		read_pg_if_n(mdb, tmp_buf, &cur_pos, name_sz);
-		mdb_unicode2ascii(mdb, tmp_buf, name_sz, pcol->name, MDB_MAX_OBJ_NAME);
+		mdb_unicode2ascii(mdb, tmp_buf, name_sz, pcol->name, sizeof(pcol->name));
 		g_free(tmp_buf);
-
-
 	}
 
 	/* Sort the columns by col_num */
@@ -317,8 +318,8 @@ GPtrArray *mdb_read_columns(MdbTableDef *table)
 		for (i=0;i<table->num_cols;i++) {
 			pcol = g_ptr_array_index(table->columns, i);
 			for (j=0; j<allprops->len; ++j) {
-				MdbProperties *props = g_array_index(allprops, MdbProperties*, j);
-				if (props->name && pcol->name && !strcmp(props->name, pcol->name)) {
+				MdbProperties *props = g_ptr_array_index(allprops, j);
+				if (props->name && !strcmp(props->name, pcol->name)) {
 					pcol->props = props;
 					break;
 				}
@@ -416,3 +417,8 @@ mdb_col_get_prop(const MdbColumn *col, const gchar *key) {
 		return NULL;
 	return g_hash_table_lookup(col->props->hash, key);
 }
+
+int mdb_col_is_shortdate(const MdbColumn *col) {
+    const char *format = mdb_col_get_prop(col, "Format");
+    return format && !strcmp(format, "Short Date");
+}
