@@ -17,18 +17,20 @@
  */
 
 #include <stdarg.h>
+#define _XOPEN_SOURCE
 #include "mdbsql.h"
 
-#ifdef DMALLOC
-#include "dmalloc.h"
-#endif
-
 #ifdef HAVE_WORDEXP_H
 #define HAVE_WORDEXP
 #include <wordexp.h>
 #endif
 
-char *g_input_ptr;
+#ifdef HAVE_STRPTIME
+#include <time.h>
+#include <stdio.h>
+#endif
+
+#include <locale.h>
 
 /* Prevent warnings from -Wmissing-prototypes.  */
 #ifdef YYPARSE_PARAM
@@ -45,8 +47,10 @@ int yyparse ();
 #endif
 #endif /* ! YYPARSE_PARAM */
 
+static MdbSargNode * mdb_sql_alloc_node(void);
+
 void
-mdb_sql_error(MdbSQL* sql, char *fmt, ...)
+mdb_sql_error(MdbSQL* sql, const char* fmt, ...)
 {
 va_list ap;
 
@@ -57,29 +61,17 @@ va_list ap;
 	fprintf(stderr, "%s\n", sql->error_msg);
 }
 
-int mdb_sql_yyinput(char *buf, int need)
-{
-int cplen, have;
-
-	have = strlen(g_input_ptr);
-	cplen = need > have ? have : need;
-	
-	if (cplen>0) {
-		memcpy(buf, g_input_ptr, cplen);
-		g_input_ptr += cplen;
-	} 
-	return cplen;
-}
 MdbSQL *mdb_sql_init()
 {
-MdbSQL *sql;
-
-	sql = (MdbSQL *) g_malloc0(sizeof(MdbSQL));
+MdbSQL *sql = g_malloc0(sizeof(MdbSQL));
 	sql->columns = g_ptr_array_new();
 	sql->tables = g_ptr_array_new();
+	sql->bound_values = g_ptr_array_new();
 	sql->sarg_tree = NULL;
 	sql->sarg_stack = NULL;
 	sql->max_rows = -1;
+	sql->limit = -1;
+	sql->limit_percent = 0;
 
 	return sql;
 }
@@ -102,14 +94,9 @@ mdb_sql_run_query (MdbSQL* sql, const gchar* querystr) {
 	g_return_val_if_fail (sql, NULL);
 	g_return_val_if_fail (querystr, NULL);
 
-	g_input_ptr = (gchar*) querystr;
-
-	/* calls to yyparse should be serialized for thread safety */
-
-	/* begin unsafe */
-	_mdb_sql (sql);
 	sql->error_msg[0]='\0';
-	if (yyparse()) {
+
+	if (parse_sql (sql, querystr)) {
 		/* end unsafe */
 		mdb_sql_error (sql, _("Could not parse '%s' command"), querystr);
 		mdb_sql_reset (sql);
@@ -123,7 +110,10 @@ mdb_sql_run_query (MdbSQL* sql, const gchar* querystr) {
 		return NULL;
 	}
 
-	mdb_sql_bind_all (sql);
+	if (mdb_sql_bind_all(sql) == -1) {
+		mdb_sql_error (sql, _("Failed to bind columns for '%s' command"), querystr);
+		return NULL;
+	}
 
 	return sql;
 }
@@ -135,7 +125,7 @@ void mdb_sql_set_maxrow(MdbSQL *sql, int maxrow)
 
 static void mdb_sql_free_columns(GPtrArray *columns)
 {
-	unsigned int i;
+	int i;
 	if (!columns) return;
 	for (i=0; i<columns->len; i++) {
 		MdbSQLColumn *c = (MdbSQLColumn *)g_ptr_array_index(columns, i);
@@ -146,7 +136,7 @@ static void mdb_sql_free_columns(GPtrArray *columns)
 }
 static void mdb_sql_free_tables(GPtrArray *tables)
 {
-	unsigned int i;
+	int i;
 	if (!tables) return;
 	for (i=0; i<tables->len; i++) {
 		MdbSQLTable *t = (MdbSQLTable *)g_ptr_array_index(tables, i);
@@ -156,6 +146,34 @@ static void mdb_sql_free_tables(GPtrArray *tables)
 	g_ptr_array_free(tables, TRUE);
 }
 
+/* This gives us a nice, uniform place to keep up with memory that needs to be freed */
+static void mdb_sql_free(MdbSQL *sql)
+{
+	/* Free MdbTableDef structures	*/
+	if (sql->cur_table) {
+		mdb_index_scan_free(sql->cur_table);
+		mdb_free_tabledef(sql->cur_table);
+		sql->cur_table = NULL;
+	}
+
+	/* Free MdbSQLColumns and MdbSQLTables */
+	mdb_sql_free_columns(sql->columns);
+	mdb_sql_free_tables(sql->tables);
+
+	/* Free sargs */
+	if (sql->sarg_tree) {
+		mdb_sql_free_tree(sql->sarg_tree);
+		sql->sarg_tree = NULL;
+	}
+
+	g_list_free(sql->sarg_stack);
+	sql->sarg_stack = NULL;
+
+	/* Free bindings  */
+	mdb_sql_unbind_all(sql);
+	g_ptr_array_free(sql->bound_values, TRUE);
+}
+
 void
 mdb_sql_close(MdbSQL *sql)
 {
@@ -173,12 +191,26 @@ MdbHandle *mdb_sql_open(MdbSQL *sql, char *db_name)
 
 #ifdef HAVE_WORDEXP
 	wordexp_t words;
+	int need_free_words = 0;
 
-	if (wordexp(db_name, &words, 0)==0) {
-		if (words.we_wordc>0) 
-			db_namep = words.we_wordv[0];
+	switch (wordexp(db_name, &words, 0))
+	{
+		case 0:
+			if (words.we_wordc>0)
+			{
+				db_namep = words.we_wordv[0];
+			}
+			need_free_words = 1;
+			break;
+		case WRDE_NOSPACE:
+			// If the error was WRDE_NOSPACE, then perhaps part of the result was allocated.
+			need_free_words = 1;
+			break;
+		default:
+			// Some other error
+			need_free_words = 0;
+			break;
 	}
-	
 #endif
 
 	sql->mdb = mdb_open(db_namep, MDB_NOFLAGS);
@@ -192,12 +224,15 @@ MdbHandle *mdb_sql_open(MdbSQL *sql, char *db_name)
 	}
 
 #ifdef HAVE_WORDEXP
-	wordfree(&words);
+	if ( need_free_words )
+	{
+		wordfree(&words);
+	}
 #endif	
 
 	return sql->mdb;
 }
-MdbSargNode *
+static MdbSargNode *
 mdb_sql_alloc_node()
 {
 	return (MdbSargNode *) g_malloc0(sizeof(MdbSargNode));
@@ -316,7 +351,7 @@ mdb_sql_dump_node(MdbSargNode *node, int level)
 			printf(" < %d\n", node->value.i); 
 			break;
 		case MDB_GT: 
-			printf(" < %d\n", node->value.i); 
+			printf(" > %d\n", node->value.i); 
 			break;
 		case MDB_LIKE: 
 			printf(" like %s\n", node->value.s); 
@@ -334,6 +369,50 @@ mdb_sql_dump_node(MdbSargNode *node, int level)
 		mdb_sql_dump_node(node->right, mylevel);
 	}
 }
+
+/* Parses date format specifier into JET date representation (double) */
+char *
+mdb_sql_strptime(MdbSQL *sql, char *data, char *format)
+{
+#ifndef HAVE_STRPTIME
+	mdb_sql_error(sql, "Your system doesn't support strptime.");
+	mdb_sql_reset(sql);
+	return NULL;
+#else
+	char *p, *pszDate;
+	struct tm tm={0};
+	double date=0;
+
+	if (data[0]!='\'' || *(p=&data[strlen(data)-1])!='\'') {
+		mdb_sql_error(sql, "First parameter of strptime (data) must be a string.");
+		mdb_sql_reset(sql);
+		return NULL;
+	}
+	*p=0; ++data;
+	if (format[0]!='\'' || *(p=&format[strlen(format)-1])!='\'') {
+		mdb_sql_error(sql, "Second parameter of strptime (format) must be a string.");
+		mdb_sql_reset(sql);
+		return NULL;
+	}
+	*p=0; ++format;
+	if (!strptime(data, format, &tm)) {
+		mdb_sql_error(sql, "strptime('%s','%s') failed.", data, format);
+		mdb_sql_reset(sql);
+		return NULL;
+	}
+	mdb_tm_to_date(&tm, &date);
+	/* It seems that when just using a time offset without date in strptime, 
+	 * we always get 1 as decimal part, een though it should be 0 for time */
+	if (date < 2 && date > 1) date--;
+	if ((pszDate=malloc(16))) {
+		char cLocale=localeconv()->decimal_point[0], *p;
+		snprintf(pszDate, 16, "%lf", date);
+		if (cLocale!='.') for (p=pszDate; *p; p++) if (*p==cLocale) *p='.';
+	}
+	return pszDate;
+#endif
+}
+
 /* evaluate a expression involving 2 constants and add answer to the stack */
 int 
 mdb_sql_eval_expr(MdbSQL *sql, char *const1, int op, char *const2)
@@ -371,7 +450,7 @@ mdb_sql_eval_expr(MdbSQL *sql, char *const1, int op, char *const2)
 		return 1;
 	}
 	if (illop) {
-		mdb_sql_error(sql, "Illegal operator used for comparision of literals.");
+		mdb_sql_error(sql, "Illegal operator used for comparison of literals.");
 		/* the column and table names are no good now */
 		mdb_sql_reset(sql);
 		return 1;
@@ -386,7 +465,7 @@ mdb_sql_eval_expr(MdbSQL *sql, char *const1, int op, char *const2)
 int 
 mdb_sql_add_sarg(MdbSQL *sql, char *col_name, int op, char *constant)
 {
-	int lastchar;
+	char *p;
 	MdbSargNode *node;
 
 	node = mdb_sql_alloc_node();
@@ -403,11 +482,15 @@ mdb_sql_add_sarg(MdbSQL *sql, char *col_name, int op, char *constant)
 	** column definition can be checked for validity
 	*/
 	if (constant[0]=='\'') {
-		lastchar = strlen(constant) > 256 ? 256 : strlen(constant);
-		strncpy(node->value.s, &constant[1], lastchar - 2);;
-		node->value.s[lastchar - 1]='\0';
+		snprintf(node->value.s, sizeof(node->value.s), "%*s", (int)strlen(constant) - 2, &constant[1]);
+		node->val_type = MDB_TEXT;
+	} else if ((p=strchr(constant, '.'))) {
+		*p=localeconv()->decimal_point[0];
+		node->value.d = strtod(constant, NULL);
+		node->val_type = MDB_DOUBLE;
 	} else {
 		node->value.i = atoi(constant);
+		node->val_type = MDB_INT;
 	}
 
 	mdb_sql_push_node(sql, node);
@@ -419,21 +502,45 @@ mdb_sql_all_columns(MdbSQL *sql)
 {
 	sql->all_columns=1;	
 }
+void
+mdb_sql_sel_count(MdbSQL *sql)
+{
+	sql->sel_count=1;
+}
+
 int mdb_sql_add_column(MdbSQL *sql, char *column_name)
 {
-	MdbSQLColumn *c;
-
-	c = (MdbSQLColumn *) g_malloc0(sizeof(MdbSQLColumn));
+	MdbSQLColumn *c = g_malloc0(sizeof(MdbSQLColumn));
 	c->name = g_strdup(column_name);
 	g_ptr_array_add(sql->columns, c);
 	sql->num_columns++;
 	return 0;
 }
+int mdb_sql_add_limit(MdbSQL *sql, char *limit, int percent)
+{
+	sql->limit = atoi(limit);
+	sql->limit_percent = percent;
+
+	if (sql->limit_percent && (sql->limit < 0 || sql->limit > 100)) {
+		return 1;
+	}
+
+	return 0;
+}
+
+int mdb_sql_get_limit(MdbSQL *sql)
+{
+	return sql->limit;
+}
+
+int mdb_sql_add_function1(MdbSQL *sql, char *func_name, char *arg1)
+{
+	fprintf(stderr, "calling function %s with %s", func_name, arg1);
+	return 0;
+}
 int mdb_sql_add_table(MdbSQL *sql, char *table_name)
 {
-	MdbSQLTable *t;
-
-	t = (MdbSQLTable *) g_malloc0(sizeof(MdbSQLTable));
+	MdbSQLTable *t = g_malloc0(sizeof(MdbSQLTable));
 	t->name = g_strdup(table_name);
 	t->alias = NULL;
 	g_ptr_array_add(sql->tables, t);
@@ -457,42 +564,33 @@ void mdb_sql_dump(MdbSQL *sql)
 }
 void mdb_sql_exit(MdbSQL *sql)
 {
-	mdb_sql_reset(sql); // Free memory
+	mdb_sql_free(sql);
 	if (sql->mdb)
 		mdb_close(sql->mdb);
+	
+	/* Cleanup the SQL engine object */
+	g_free(sql);
 }
 void mdb_sql_reset(MdbSQL *sql)
 {
-	if (sql->cur_table) {
-		mdb_index_scan_free(sql->cur_table);
-		if (sql->cur_table->sarg_tree) {
-			mdb_sql_free_tree(sql->cur_table->sarg_tree);
-			sql->cur_table->sarg_tree = NULL;
-		}
-		mdb_free_tabledef(sql->cur_table);
-		sql->cur_table = NULL;
-	}
+	mdb_sql_free(sql);
 
 	/* Reset columns */
-	mdb_sql_free_columns(sql->columns);
 	sql->num_columns = 0;
 	sql->columns = g_ptr_array_new();
 
-	/* Reset tables */
-	mdb_sql_free_tables(sql->tables);
+	/* Reset MdbSQL tables */
 	sql->num_tables = 0;
 	sql->tables = g_ptr_array_new();
 
-	/* Reset sargs */
-	if (sql->sarg_tree) {
-		mdb_sql_free_tree(sql->sarg_tree);
-		sql->sarg_tree = NULL;
-	}
-	g_list_free(sql->sarg_stack);
-	sql->sarg_stack = NULL;
+	/* Reset bindings */
+	sql->bound_values = g_ptr_array_new();
 
 	sql->all_columns = 0;
+	sql->sel_count = 0;
 	sql->max_rows = -1;
+	sql->row_count = 0;
+	sql->limit = -1;
 }
 static void print_break(int sz, int first)
 {
@@ -545,7 +643,7 @@ void mdb_sql_listtables(MdbSQL *sql)
  		entry = g_ptr_array_index (mdb->catalog, i);
  		if (mdb_is_user_table(entry)) {
 			//col = g_ptr_array_index(table->columns,0);
-			tmpsiz = mdb_ascii2unicode(mdb, entry->object_name, 0, tmpstr, 100);
+			tmpsiz = mdb_ascii2unicode(mdb, entry->object_name, 0, tmpstr, sizeof(tmpstr));
 			mdb_fill_temp_field(&fields[0],tmpstr, tmpsiz, 0,0,0,0);
 			row_size = mdb_pack_row(ttable, row_buffer, 1, fields);
 			mdb_add_row_to_pg(ttable,row_buffer, row_size);
@@ -609,15 +707,15 @@ void mdb_sql_describe_table(MdbSQL *sql)
 	for (i=0;i<table->num_cols;i++) {
 
 		col = g_ptr_array_index(table->columns,i);
-		tmpsiz = mdb_ascii2unicode(mdb, col->name, 0, col_name, 100);
+		tmpsiz = mdb_ascii2unicode(mdb, col->name, 0, col_name, sizeof(col_name));
 		mdb_fill_temp_field(&fields[0],col_name, tmpsiz, 0,0,0,0);
 
-		strcpy(tmpstr, mdb_get_colbacktype_string(col));
-		tmpsiz = mdb_ascii2unicode(mdb, tmpstr, 0, col_type, 100);
+		snprintf(tmpstr, sizeof(tmpstr), "%s", mdb_get_colbacktype_string(col));
+		tmpsiz = mdb_ascii2unicode(mdb, tmpstr, 0, col_type, sizeof(col_type));
 		mdb_fill_temp_field(&fields[1],col_type, tmpsiz, 0,0,0,1);
 
-		sprintf(tmpstr,"%d",col->col_size);
-		tmpsiz = mdb_ascii2unicode(mdb, tmpstr, 0, col_size, 100);
+		snprintf(tmpstr, sizeof(tmpstr), "%d", col->col_size);
+		tmpsiz = mdb_ascii2unicode(mdb, tmpstr, 0, col_size, sizeof(col_size));
 		mdb_fill_temp_field(&fields[2],col_size, tmpsiz, 0,0,0,2);
 
 		row_size = mdb_pack_row(ttable, row_buffer, 3, fields);
@@ -630,24 +728,46 @@ void mdb_sql_describe_table(MdbSQL *sql)
 	sql->cur_table = ttable;
 }
 
+MdbColumn *mdb_sql_find_colbyname(MdbTableDef *table, char *name)
+{
+	unsigned int i;
+	MdbColumn *col;
+
+	for (i=0;i<table->num_cols;i++) {
+		col=g_ptr_array_index(table->columns,i);
+		if (!g_ascii_strcasecmp(col->name, name)) return col;
+	}
+	return NULL;
+}
+
 int mdb_sql_find_sargcol(MdbSargNode *node, gpointer data)
 {
 	MdbTableDef *table = data;
-	unsigned int i;
 	MdbColumn *col;
 
 	if (!mdb_is_relational_op(node->op)) return 0;
 	if (!node->parent) return 0;
 
-	for (i=0;i<table->num_cols;i++) {
-		col=g_ptr_array_index(table->columns,i);
-		if (!strcasecmp(col->name, (char *)node->parent)) {
-			node->col = col;
-			break;
+	if ((col = mdb_sql_find_colbyname(table, (char *)node->parent))) {
+		node->col = col;
+		/* Do conversion to required target value type.
+		 * Plain integers are UNIX timestamps for backwards compatibility of parser
+		*/
+		if (col->col_type == MDB_DATETIME && node->val_type == MDB_INT) {
+			struct tm *tmp;
+#ifdef HAVE_GMTIME_R
+			struct tm tm;
+			tmp = gmtime_r((time_t*)&node->value.i, &tm);
+#else		// regular gmtime on Windows uses thread-local storage
+			tmp = gmtime((time_t*)&node->value.i);
+#endif
+			mdb_tm_to_date(tmp, &node->value.d);
+			node->val_type = MDB_DOUBLE;
 		}
 	}
 	return 0;
 }
+
 void 
 mdb_sql_select(MdbSQL *sql)
 {
@@ -664,6 +784,7 @@ int found = 0;
 		return;
 	}
 
+	if (!sql->num_tables) return;
 	sql_tab = g_ptr_array_index(sql->tables,0);
 
 	table = mdb_read_table_by_name(mdb, sql_tab->name, MDB_TABLE);
@@ -674,6 +795,27 @@ int found = 0;
 		return;
 	}
 	mdb_read_columns(table);
+
+	if (sql->sel_count && !sql->sarg_tree) {
+		MdbTableDef *ttable = mdb_create_temp_table(mdb, "#count");
+		char tmpstr[32];
+		gchar row_cnt[32];
+		unsigned char row_buffer[MDB_PGSIZE];
+		MdbField fields[1];
+		int row_size, tmpsiz;
+
+		mdb_sql_add_temp_col(sql, ttable, 0, "count", MDB_TEXT, 30, 0);
+		snprintf(tmpstr, sizeof(tmpstr), "%d", table->num_rows);
+		tmpsiz = mdb_ascii2unicode(mdb, tmpstr, 0, row_cnt, sizeof(row_cnt));
+		mdb_fill_temp_field(&fields[0],row_cnt, tmpsiz, 0,0,0,0);
+		row_size = mdb_pack_row(ttable, row_buffer, 1, fields);
+		mdb_add_row_to_pg(ttable,row_buffer, row_size);
+		ttable->num_rows++;
+		sql->cur_table = ttable;
+		mdb_free_tabledef(table);
+		return;
+	}
+
 	mdb_read_indices(table);
 	mdb_rewind_table(table);
 
@@ -689,7 +831,7 @@ int found = 0;
 		found=0;
 		for (j=0;j<table->num_cols;j++) {
 			col=g_ptr_array_index(table->columns,j);
-			if (!strcasecmp(sqlcol->name, col->name)) {
+			if (!g_ascii_strcasecmp(sqlcol->name, col->name)) {
 				sqlcol->disp_size = mdb_col_disp_size(col);
 				found=1;
 				break;
@@ -697,6 +839,8 @@ int found = 0;
 		}
 		if (!found) {
 			mdb_sql_error(sql, "Column %s not found",sqlcol->name);
+			mdb_index_scan_free(table);
+			mdb_free_tabledef(table);
 			mdb_sql_reset(sql);
 			return;
 		}
@@ -718,27 +862,54 @@ int found = 0;
 
 	sql->cur_table = table;
 	mdb_index_scan_init(mdb, table);
+
+	/* We know how many rows there are, so convert limit percentage
+	 * to an row count */
+	if (sql->limit != -1 && sql->limit_percent) {
+		sql->limit = (int)((double)table->num_rows / 100 * sql->limit);
+		sql->limit_percent = 0;
+	}
 }
 
-void 
+int
 mdb_sql_bind_column(MdbSQL *sql, int colnum, void *varaddr, int *len_ptr)
 {
 	MdbSQLColumn *sqlcol;
 
+	if (colnum <= 0 || colnum > sql->num_columns)
+		return -1;
+
 	/* sql columns are traditionally 1 based, so decrement colnum */
 	sqlcol = g_ptr_array_index(sql->columns,colnum - 1);
-	mdb_bind_column_by_name(sql->cur_table, sqlcol->name, varaddr, len_ptr);
+	return mdb_bind_column_by_name(sql->cur_table, sqlcol->name, varaddr, len_ptr);
 }
-void 
+
+int
 mdb_sql_bind_all(MdbSQL *sql)
 {
 	unsigned int i;
+	void *bound_value;
 
 	for (i=0;i<sql->num_columns;i++) {
-		sql->bound_values[i] = g_malloc0(MDB_BIND_SIZE);
-		mdb_sql_bind_column(sql, i+1, sql->bound_values[i], NULL);
+		bound_value = g_malloc0(sql->mdb->bind_size);
+		g_ptr_array_add(sql->bound_values, bound_value);
+		if (mdb_sql_bind_column(sql, i+1, bound_value, NULL) == -1) {
+			mdb_sql_unbind_all(sql);
+			return -1;
+		}
+	}
+	return sql->num_columns;
+}
+
+void mdb_sql_unbind_all(MdbSQL *sql)
+{
+	unsigned int i;
+
+	for (i=0;i<sql->bound_values->len;i++) {
+		g_free(g_ptr_array_index(sql->bound_values, i));
 	}
 }
+
 /*
  * mdb_sql_fetch_row is now just a wrapper around mdb_fetch_row.
  *   It is left here only for backward compatibility.
@@ -746,7 +917,14 @@ mdb_sql_bind_all(MdbSQL *sql)
 int
 mdb_sql_fetch_row(MdbSQL *sql, MdbTableDef *table)
 {
-	return mdb_fetch_row(table);
+	int rc = mdb_fetch_row(table);
+	if (rc) {
+		if (sql->limit >= 0 && sql->row_count + 1 > sql->limit) {
+			return 0;
+		}
+		sql->row_count++;
+	}
+	return rc;
 }
 
 void 
@@ -776,7 +954,7 @@ mdb_sql_dump_results(MdbSQL *sql)
 	while(mdb_fetch_row(sql->cur_table)) {
   		for (j=0;j<sql->num_columns;j++) {
 			sqlcol = g_ptr_array_index(sql->columns,j);
-			print_value(sql->bound_values[j],sqlcol->disp_size,!j);
+			print_value(g_ptr_array_index(sql->bound_values, j),sqlcol->disp_size,!j);
 		}
 		fprintf(stdout,"\n");
 	}
@@ -788,10 +966,6 @@ mdb_sql_dump_results(MdbSQL *sql)
 	}
 
 	fprintf(stdout,"\n");
-	/* clean up */
-	for (j=0;j<sql->num_columns;j++) {
-		g_free(sql->bound_values[j]);
-	}
 
 	/* the column and table names are no good now */
 	mdb_sql_reset(sql);
