@@ -47,10 +47,6 @@ extern void clear_history ();
 #include <string.h>
 #include "mdbsql.h"
 
-#ifdef DMALLOC
-#include "dmalloc.h"
-#endif
-
 void dump_results(FILE *out, MdbSQL *sql, char *delimiter);
 void dump_results_pp(FILE *out, MdbSQL *sql);
 
@@ -69,11 +65,11 @@ int noexec = 0;
 #ifndef HAVE_LIBREADLINE
 char *readline(char *prompt)
 {
-char *buf, line[1000];
+char line[1000];
 int i = 0;
 
-	puts(prompt);
-	if (! fgets(line,1000,stdin)) {
+	fputs(prompt, stdout);
+	if (!fgets(line, sizeof(line), stdin)) {
 		return NULL;
 	}
 	for (i=strlen(line);i>0;i--) {
@@ -82,10 +78,8 @@ int i = 0;
 			break;
 		}
 	}
-	buf = (char *) malloc(strlen(line)+1);
-	strcpy(buf,line);
 
-	return buf;
+	return g_strdup(line);
 }
 #endif
 
@@ -156,40 +150,6 @@ do_set_cmd(MdbSQL *sql, char *s)
 	}
 }
 
-int
-read_file(char *s, int line, unsigned int *bufsz, char *mybuf)
-{
-	char *fname;
-	FILE *in;
-	char buf[256];
-	unsigned int cursz = 0;	
-	int lines = 0;	
-
-	fname = s;
-	while (*fname && *fname==' ') fname++;
-
-	if (! (in = fopen(fname, "r"))) {
-		fprintf(stderr,"Unable to open file %s\n", fname);
-		mybuf[0]=0;
-		return 0;
-	}
-	while (fgets(buf, 255, in)) {
-		cursz += strlen(buf) + 1;
-		if (cursz > (*bufsz)) {
-			(*bufsz) *= 2;
-			mybuf = (char *) realloc(mybuf, *bufsz);
-		}	
-		strcat(mybuf, buf);
-#ifdef HAVE_READLINE_HISTORY
-		/* don't record blank lines */
-		if (strlen(buf)) add_history(buf);
-#endif
-		strcat(mybuf, "\n");
-		lines++;
-		printf("%d => %s",line+lines, buf);
-	}
-	return lines;
-}
 void 
 run_query(FILE *out, MdbSQL *sql, char *mybuf, char *delimiter)
 {
@@ -205,14 +165,14 @@ run_query(FILE *out, MdbSQL *sql, char *mybuf, char *delimiter)
 			else 
 				printf("Index scanning %s using %s\n", table->name, table->scan_idx->name);
 		}
-		if (noexec) {
-			mdb_sql_reset(sql);
-			return;
+		/* If noexec != on, dump results */
+		if (!noexec) {
+			if (pretty_print)
+				dump_results_pp(out, sql);
+			else
+				dump_results(out, sql, delimiter);
 		}
-		if (pretty_print)
-			dump_results_pp(out, sql);
-		else
-			dump_results(out, sql, delimiter);
+		mdb_sql_reset(sql);
 	}
 }
 
@@ -253,7 +213,6 @@ dump_results(FILE *out, MdbSQL *sql, char *delimiter)
 {
 	unsigned int j;
 	MdbSQLColumn *sqlcol;
-	unsigned long row_count = 0;
 
 	if (headers) {
 		for (j=0;j<sql->num_columns-1;j++) {
@@ -266,23 +225,20 @@ dump_results(FILE *out, MdbSQL *sql, char *delimiter)
 		fprintf(out,"\n");
 		fflush(out);
 	}
-	while(mdb_fetch_row(sql->cur_table)) {
-		row_count++;
+	while(mdb_sql_fetch_row(sql, sql->cur_table)) {
   		for (j=0;j<sql->num_columns-1;j++) {
 			sqlcol = g_ptr_array_index(sql->columns,j);
-			fprintf(out, "%s%s", (char*)(sql->bound_values[j]),
+			fprintf(out, "%s%s", (char*)(g_ptr_array_index(sql->bound_values, j)),
 				delimiter ? delimiter : "\t");
 		}
 		sqlcol = g_ptr_array_index(sql->columns,sql->num_columns-1);
-		fprintf(out, "%s", (char*)(sql->bound_values[sql->num_columns-1]));
+		fprintf(out, "%s", (char*)(g_ptr_array_index(sql->bound_values, sql->num_columns-1)));
 		fprintf(out,"\n");
 		fflush(out);
 	}
 	if (footers) {
-		print_rows_retrieved(out, row_count);
+		print_rows_retrieved(out, sql->row_count);
 	}
-
-	mdb_sql_reset(sql);
 }
 
 void 
@@ -290,13 +246,12 @@ dump_results_pp(FILE *out, MdbSQL *sql)
 {
 	unsigned int j;
 	MdbSQLColumn *sqlcol;
-	unsigned long row_count = 0;
 
 	/* print header */
 	if (headers) {
 		for (j=0;j<sql->num_columns;j++) {
 			sqlcol = g_ptr_array_index(sql->columns,j);
-			if (strlen(sqlcol->name)>sqlcol->disp_size)
+			if (strlen(sqlcol->name)>(size_t)sqlcol->disp_size)
 				sqlcol->disp_size = strlen(sqlcol->name);
 			print_break(out, sqlcol->disp_size, !j);
 		}
@@ -318,11 +273,10 @@ dump_results_pp(FILE *out, MdbSQL *sql)
 	fflush(out);
 
 	/* print each row */
-	while(mdb_fetch_row(sql->cur_table)) {
-		row_count++;
+	while(mdb_sql_fetch_row(sql, sql->cur_table)) {
   		for (j=0;j<sql->num_columns;j++) {
 			sqlcol = g_ptr_array_index(sql->columns,j);
-			print_value(out, sql->bound_values[j],sqlcol->disp_size,!j);
+			print_value(out, (char *) g_ptr_array_index(sql->bound_values, j), sqlcol->disp_size,!j);
 		}
 		fprintf(out,"\n");
 		fflush(out);
@@ -336,33 +290,79 @@ dump_results_pp(FILE *out, MdbSQL *sql)
 	fprintf(out,"\n");
 	fflush(out);
 	if (footers) {
-		print_rows_retrieved(out, row_count);
+		print_rows_retrieved(out, sql->row_count);
+	}
+}
+
+static char *
+find_sql_terminator(char *s)
+{
+	char *sp;
+	int len = strlen(s);
+
+	if (len == 0) {
+		return NULL;
 	}
 
-	/* clean up */
-	for (j=0;j<sql->num_columns;j++) {
-		g_free(sql->bound_values[j]);
+	sp = &s[len-1];
+	while (sp > s && isspace(*sp)) {
+		sp--;
 	}
 
-	mdb_sql_reset(sql);
+	if (*sp == ';') {
+		return sp;
+	}
+
+	return NULL;
 }
 
 int
 main(int argc, char **argv)
 {
-char *s = NULL;
-char prompt[20];
-int line = 0;
-char *mybuf;
-unsigned int bufsz;
-MdbSQL *sql;
-int opt;
-FILE *in = NULL, *out = NULL;
-char *home = getenv("HOME");
-char *histpath;
-char *delimiter = NULL;
+	char *s = NULL;
+	char prompt[20];
+	int line = 0;
+	char *mybuf;
+	unsigned int bufsz;
+	MdbSQL *sql;
+	FILE *in = NULL, *out = NULL;
+	char *filename_in=NULL, *filename_out=NULL;
+#ifdef HAVE_READLINE_HISTORY
+	char *home = getenv("HOME");
+	char *histpath;
+#endif
+	char *delimiter = NULL;
+	int in_from_colon_r = 0;
 
 
+	GOptionEntry entries[] = {
+		{ "delim", 'd', 0, G_OPTION_ARG_STRING, &delimiter, "Use this delimiter.", "char"},
+		{ "no-pretty-print", 'P', G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, &pretty_print, "Don't pretty print", NULL},
+		{ "no-header", 'H', G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, &headers, "Don't print header", NULL},
+		{ "no-footer", 'F', G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, &footers, "Don't print footer", NULL},
+		{ "input", 'i', 0, G_OPTION_ARG_STRING, &filename_in, "Read SQL from specified file", "file"},
+		{ "output", 'o', 0, G_OPTION_ARG_STRING, &filename_out, "Write result to specified file", "file"},
+		{ NULL },
+	};
+	GError *error = NULL;
+	GOptionContext *opt_context;
+
+	opt_context = g_option_context_new("<file> - Run SQL");
+	g_option_context_add_main_entries(opt_context, entries, NULL /*i18n*/);
+	// g_option_context_set_strict_posix(opt_context, TRUE); /* options first, requires glib 2.44 */
+	if (!g_option_context_parse (opt_context, &argc, &argv, &error))
+	{
+		fprintf(stderr, "option parsing failed: %s\n", error->message);
+		fputs(g_option_context_get_help(opt_context, TRUE, NULL), stderr);
+		exit (1);
+	}
+
+	if (argc > 2) {
+		fputs("Wrong number of arguments.\n\n", stderr);
+		fputs(g_option_context_get_help(opt_context, TRUE, NULL), stderr);
+		exit(1);
+	}
+
 #ifdef HAVE_READLINE_HISTORY
 	if (home) {
 		histpath = (char *) g_strconcat(home, "/", HISTFILE, NULL);
@@ -374,49 +374,31 @@ char *delimiter = NULL;
 	if (!isatty(fileno(stdin))) {
 		in = stdin;
 	}
-
-	while ((opt=getopt(argc, argv, "HFpd:i:o:"))!=-1) {
-		switch (opt) {
-		        case 'd':
-				delimiter = (char *) g_strdup(optarg);
-				break;
-		        case 'p':
-				pretty_print=0;
-				break;
-		        case 'H':
-				headers=0;
-				break;
-		        case 'F':
-				footers=0;
-				break;
-		        case 'i':
-				if (!strcmp(optarg, "stdin"))
-					in = stdin;
-				else if (!(in = fopen(optarg, "r"))) {
-					fprintf(stderr,"Unable to open file %s\n", optarg);
-					exit(1);
-				}
-				break;
-		        case 'o':
-				if (!(out = fopen(optarg, "w"))) {
-					fprintf(stderr,"Unable to open file %s\n", optarg);
-					exit(1);
-				}
-				break;
-			default:
-				fprintf(stdout,"Unknown option.\nUsage: %s [-HFp] [-d <delimiter>] [-i <file>] [-o <file>] [<database>]\n", argv[0]);
-				exit(1);
+	if (filename_in) {
+		if (!strcmp(filename_in, "stdin"))
+			in = stdin;
+		else if (!(in = fopen(filename_in, "r"))) {
+			fprintf(stderr, "Unable to open file %s\n", filename_in);
+			exit(1);
 		}
 	}
+	if (filename_out) {
+		if (!(out = fopen(filename_out, "w"))) {
+			fprintf(stderr,"Unable to open file %s\n", filename_out);
+			exit(1);
+		}
+	}
+
+
 	/* initialize the SQL engine */
 	sql = mdb_sql_init();
-	if (argc>optind) {
-		mdb_sql_open(sql, argv[optind]);
+	if (argc == 2) {
+		mdb_sql_open(sql, argv[1]);
 	}
 
 	/* give the buffer an initial size */
 	bufsz = 4096;
-	mybuf = (char *) g_malloc(bufsz);
+	mybuf = g_malloc(bufsz);
 	mybuf[0]='\0';
 
 	while (1) {
@@ -424,18 +406,24 @@ char *delimiter = NULL;
 		if (s) free(s);
 
 		if (in) {
-			s=malloc(256);
-			if ((!s) || (!fgets(s, 256, in))) {
-				/* if we have something in the buffer, run it */
-				if (strlen(mybuf))
-					run_query((out) ? out : stdout,
-					          sql, mybuf, delimiter);
-				break;
-			}
-			if (s[strlen(s)-1]=='\n')
+			s=calloc(bufsz, 1);
+			if (!fgets(s, bufsz, in)) {
+				// Backwards compatibility with older MDBTools
+				// Files read from the command line had an
+				// implicit "go" at the end
+				if (!in_from_colon_r && strlen(mybuf))
+					strcpy(s, "go");
+				else if (in_from_colon_r) {
+					line = 0;
+					fclose(in);
+					in = NULL;
+					in_from_colon_r = 0;
+				} else
+					break;
+			} else if (s[strlen(s)-1]=='\n')
 				s[strlen(s)-1]=0;
 		} else {
-			sprintf(prompt, "%d => ", line);
+			snprintf(prompt, sizeof(prompt), "%d => ", line);
 			s=readline(prompt);
 			if (!s)
 				break;
@@ -455,25 +443,47 @@ char *delimiter = NULL;
 			line = 0;
 			mybuf[0]='\0';
 		} else if (!strncmp(s,":r",2)) {
-			line += read_file(&s[2], line, &bufsz, mybuf);
+			char *fname = &s[2];
+			if (in) {
+				fprintf(stderr, "Can not handle nested opens\n");
+			} else {
+				while (*fname && isspace(*fname))
+					fname++;
+				if (!(in = fopen(fname, "r"))) {
+					fprintf(stderr,"Unable to open file %s\n", fname);
+					mybuf[0]=0;
+				} else {
+					in_from_colon_r = 1;
+				}
+			}
 		} else {
+			char *p;
+
 			while (strlen(mybuf) + strlen(s) > bufsz) {
 				bufsz *= 2;
 				mybuf = (char *) g_realloc(mybuf, bufsz);
 			}
 #ifdef HAVE_READLINE_HISTORY
-			/* don't record blank lines */
-			if (strlen(s)) add_history(s);
+			/* don't record blank lines, or lines read from files
+			 * specified on the command line */
+			if ((!in || in_from_colon_r) && strlen(s))
+				add_history(s);
 #endif
 			strcat(mybuf,s);
 			/* preserve line numbering for the parser */
 			strcat(mybuf,"\n");
+
+			if ((p = find_sql_terminator(mybuf))) {
+				*p = '\0';
+				line = 0;
+				run_query((out) ? out : stdout, sql, mybuf, delimiter);
+				mybuf[0]='\0';
+			}
 		}
 	}
-	mdb_sql_exit(sql);	
+	mdb_sql_exit(sql);
 
 	g_free(mybuf);
-	g_free(delimiter);
 	if (s) free(s);
 	if (out) fclose(out);
 	if ((in) && (in != stdin)) fclose(in);
@@ -487,6 +497,11 @@ char *delimiter = NULL;
 	}
 #endif
 
+	g_option_context_free(opt_context);
+	g_free(delimiter);
+	g_free(filename_in);
+	g_free(filename_out);
+
 	return 0;
 }
 #else
